from typing import Any, Union, List, Optional, Dict

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util.errors import MissingDependencyError, BackendError


class GurobiBackend(Backend):
    """`Backend` implementation for the Gurobi Solver."""

    def __init__(self,
                 time_limit: Optional[float] = None,
                 solution_limit: Optional[float] = None,
                 verbose: bool = False,
                 **solver_args):
        """
        :param time_limit:
            The solver time limit.

        :param solution_limit:
            The solver solution limit.

        :param verbose:
            Whether or not to print information during the optimization process.

        :param solver_args:
            Parameters of the solver to be set via the `model.SetParam()` function.
        """
        super(GurobiBackend, self).__init__()

        try:
            import gurobipy
            self._gp = gurobipy
            """The lazily imported gurobipy instance."""
        except ModuleNotFoundError:
            raise MissingDependencyError(package='gurobipy')

        self.verbose: bool = verbose
        """Whether or not to print information during the optimization process."""

        self.solver_args: Dict[str, Any] = solver_args
        """Parameters of the solver to be set via the `model.SetParam()` function."""

        self._env: Optional = None
        """The gurobi environment instance."""

        if time_limit is not None and 'TimeLimit' not in self.solver_args:
            self.solver_args['TimeLimit'] = time_limit
        if solution_limit is not None and 'SolutionLimit' not in self.solver_args:
            self.solver_args['SolutionLimit'] = solution_limit

    def _build_model(self) -> Any:
        self._env = self._gp.Env(empty=True)
        self._env.setParam('OutputFlag', self.verbose)
        self._env.start()
        model = self._gp.Model(env=self._env, name='model')
        for param, value in self.solver_args.items():
            model.setParam(param, value)
        return model

    def _solve_model(self) -> Optional:
        self.model.update()
        self.model.optimize()
        return None if self.model.SolCount == 0 else self.model

    def clear(self) -> Any:
        self._env.dispose()
        self.model.dispose()
        super(GurobiBackend, self).clear()

    def minimize(self, cost) -> Any:
        self.model.setObjective(cost, self._gp.GRB.MINIMIZE)
        return self

    def maximize(self, cost) -> Any:
        self.model.setObjective(cost, self._gp.GRB.MAXIMIZE)
        return self

    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
        self.model.addConstrs((c for c in constraints), name=name)
        return self

    def add_variable(self, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> Any:
        if not hasattr(self._gp.GRB, vtype.upper()):
            raise BackendError(unsupported=f"vtype '{vtype}'")
        # addVar does not accept name=None as parameter
        kwargs = dict() if name is None else dict(name=name)
        var = self.model.addVar(vtype=getattr(self._gp.GRB, vtype.upper()), lb=lb, ub=ub, **kwargs)
        return var

    def add_variables(self, *keys: int, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> np.ndarray:
        if not hasattr(self._gp.GRB, vtype.upper()):
            raise BackendError(unsupported=f"vtype '{vtype}'")
        vtype = getattr(self._gp.GRB, vtype.upper())
        var = self.model.addVars(*keys, vtype=vtype, lb=lb, ub=ub, name=name).values()
        return np.array(var).reshape(keys)

    def get_objective(self) -> float:
        return self.solution.objVal

    def get_values(self, expressions: np.ndarray) -> np.ndarray:
        values = [v.x if isinstance(v, self._gp.Var) else v.getValue() for v in expressions.flatten()]
        return np.reshape(values, expressions.shape)

    def sum(self, a: np.ndarray, aux: Optional[str] = None) -> Any:
        return self.aux(expressions=np.sum(a), aux_vtype=aux)

    def square(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        return self.aux(expressions=a ** 2, aux_vtype=aux)

    def sqrt(self, a: np.ndarray, aux: Optional[str] = 'continuous') -> np.ndarray:
        self._aux_warning(exp='continuous', aux=aux, msg='to compute square roots')
        # creating auxiliary variables is necessary since 'addGenConstrPow' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        sqrt_vector = self.add_continuous_variables(*a.shape, lb=0.0, ub=float('inf'))
        for aux_var, sqrt_var in zip(aux_vector.flatten(), sqrt_vector.flatten()):
            self.model.addGenConstrPow(sqrt_var, aux_var, 2)
        return sqrt_vector

    def abs(self, a: np.ndarray, aux: Optional[str] = 'continuous') -> np.ndarray:
        self._aux_warning(exp='continuous', aux=aux, msg='to compute absolute values')
        # creating auxiliary variables is necessary since 'addGenConstrAbs' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        abs_vector = self.add_continuous_variables(*a.shape, lb=0.0, ub=float('inf'))
        for aux_var, abs_var in zip(aux_vector.flatten(), abs_vector.flatten()):
            self.model.addGenConstrAbs(abs_var, aux_var)
        return abs_vector

    def log(self, a: np.ndarray, aux: Optional[str] = 'continuous') -> np.ndarray:
        self._aux_warning(exp='continuous', aux=aux, msg='to compute logarithms')
        # creating auxiliary variables is necessary since 'addGenConstrExp' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        log_vector = self.add_continuous_variables(*a.shape, lb=-float('inf'), ub=float('inf'))
        for aux_var, log_var in zip(aux_vector.flatten(), log_vector.flatten()):
            self.model.addGenConstrExp(log_var, aux_var)
        return log_vector

    def divide(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
        try:
            return super(GurobiBackend, self).divide(a, b, aux=aux)
        except self._gp.GurobiError:
            # gurobipy.GurobiError: Divisor must be a constant
            # in case the divisor is not an array of constants, we handle the case by adding N auxiliary variables z_i
            # so that a_i / b_i = z_i --> a_i = z_i * b_i
            self._aux_warning(exp='continuous', aux=aux, msg='to compute divisions with non-constant divisors')
            z = self.add_continuous_variables(*a.shape)
            self.add_constraints([ai == zi * bi for ai, bi, zi in zip(a.flatten(), b.flatten(), z.flatten())])
            return z



"""Interfaces and classes for Moving Targets Learners."""

from moving_targets.learners.learner import Learner
from moving_targets.learners.scikit_learners import ScikitLearner, ScikitClassifier, LinearRegression, \
    LogisticRegression, RandomForestRegressor, RandomForestClassifier, GradientBoostingRegressor, \
    GradientBoostingClassifier
from moving_targets.learners.tensorflow_learners import TensorflowLearner, MultiLayerPerceptron



"""Moving Targets Optimizers for Alpha/Beta Hyper-parameters in Master."""

from moving_targets.masters.optimizers.optimizer import Optimizer
from moving_targets.masters.optimizers.policies import ConstantSlope, ExponentialSlope
from moving_targets.masters.optimizers.satisfiability import BetaSatisfiability, BetaClassSatisfiability, \
    BetaBoundedSatisfiability



from typing import Any, Union, List, Optional

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util.errors import BackendError


class NumpyBackend(Backend):
    """`Backend` implementation for Numpy.

    Numpy cannot be used as a solver, still, this backend may come in handy when a loss object is needed to compute the
    loss between two numeric vectors, instead of a numeric vector and a vector of model variables.
    """

    _ERROR_MESSAGE: str = 'Numpy is not a solver, thus '
    """Error message for unsupported operations."""

    def __init__(self, clip: Optional[float] = 1e-15):
        """
        :param clip:
            The clipping value to use when calling the 'log()' function to avoid numeric errors. If None, does not clip.
        """
        super(NumpyBackend, self).__init__()

        self.clip: float = clip
        """The clipping value to use when calling the 'log()' function to avoid numeric errors."""

    def _build_model(self) -> Any:
        return np

    def _solve_model(self) -> Optional:
        return np

    def minimize(self, cost) -> Any:
        raise BackendError(unsupported='no solution can be minimized', message=self._ERROR_MESSAGE)

    def maximize(self, cost) -> Any:
        raise BackendError(unsupported='no solution can be maximized', message=self._ERROR_MESSAGE)

    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
        raise BackendError(unsupported='no constraint can be added')

    def add_variable(self, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> Any:
        raise BackendError(unsupported='no variable can be added', message=self._ERROR_MESSAGE)

    def get_objective(self) -> float:
        raise BackendError(unsupported='no objective can be retrieved', message=self._ERROR_MESSAGE)

    def add_constant(self, val: Any, vtype: str = 'continuous', name: Optional[str] = None) -> Any:
        return val

    def add_constants(self, val: np.ndarray, vtype: str = 'continuous', name: Optional[str] = None) -> np.ndarray:
        return val

    def get_values(self, expressions: np.ndarray) -> np.ndarray:
        return expressions

    def aux(self,
            expressions: Any,
            aux_vtype: Optional[str] = None,
            aux_lb: float = -float('inf'),
            aux_ub: float = float('inf'),
            aux_name: Optional[str] = None) -> Any:
        self._aux_warning(exp=None, aux=aux_vtype, msg='to compute any value')
        return expressions

    def sum(self, a: np.ndarray, aux: Optional[str] = None) -> Any:
        return np.sum(a)

    def square(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        return np.square(a)

    def sqrt(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        return np.sqrt(a)

    def abs(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        return np.abs(a)

    def log(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        if self.clip is not None:
            a_min = np.min(a)
            if a_min < 0:
                self._LOGGER.warning(f'Trying to compute logarithm of negative number, clipping to {self.clip}')
            elif a_min < self.clip:
                self._LOGGER.info(f'Values in the interval [0, {self.clip}] will be clipped to {self.clip}')
            a = a.clip(min=self.clip)
        return np.log(a)



"""Constraints Metrics."""

from typing import Optional, Callable, List, Union, Dict

import numpy as np
import pandas as pd

from moving_targets.metrics.metric import Metric
from moving_targets.util import probabilities


class ClassFrequenciesStd(Metric):
    """Standard Deviation of the Class Frequencies, usually constrained to be null."""

    def __init__(self, classes: Optional[int] = None, name: str = 'std'):
        """
        :param classes:
            The number of classes or None for automatic class inference.

        :param name:
            The name of the metric.
        """
        super(ClassFrequenciesStd, self).__init__(name=name)

        self.classes: Optional[int] = classes
        """The number of classes or None for automatic class inference."""

    def __call__(self, x, y: np.ndarray, p: np.ndarray) -> float:
        # bincount is similar to np.unique(..., return_counts=True) but allows to fix a minimum number of classes
        # in this way, if the predictions are all the same, the counts will be [n, 0, ..., 0] instead of [n]
        n_classes = self.classes or probabilities.count_classes(vector=y)
        classes_counts = np.bincount(probabilities.get_classes(p), minlength=n_classes) / len(p)
        return classes_counts.std()


class DIDI(Metric):
    """Disparate Impact Discrimination Index."""

    @staticmethod
    def get_indicator_matrix(x: pd.DataFrame, protected: str) -> np.array:
        """Computes the indicator matrix given the input data and a protected feature.

        :param x:
            The input data (it must be a pandas dataframe because features are searched by column name).

        :param protected:
            The name of the protected feature.

            During the solving process, the algorithm checks inside the data for column names starting with the given
            feature name: if a single column is found this is assumed to be a categorical column (thus, the number of
            classes is inferred from the unique values in the column), otherwise, if multiple columns are found this is
            assumed to be a one-hot encoded version of the categorical column (thus, the number of classes is inferred
            from the number of columns).

        :return:
            The indicator matrix, i.e., a matrix in which the i-th row represents a boolean vector stating whether or
            not the j-th sample (represented by the j-th column) is part of the i-th protected group.

        :raise `AssertionError`:
            If there is no column name starting with the `protected` string.
        """
        features = [c for c in x.columns if c.startswith(protected)]
        assert len(features) > 0, AssertionError(f"No column starting with the given protected feature '{protected}'")
        matrix = x[features].values.astype(int).squeeze()
        # if a single protected column is found out, this is interpreted as a categorical column thus is onehot encoded,
        # otherwise, if multiple protected columns are found out, these are interpreted as already encoded columns
        matrix = probabilities.get_onehot(matrix, onehot_binary=True) if matrix.ndim == 1 else matrix
        return matrix.transpose().astype(bool)

    @staticmethod
    def classification_didi(indicator_matrix: np.ndarray, targets: np.ndarray):
        """Computes the Disparate Impact Discrimination Index for Classification Tasks.

        :param indicator_matrix:
            A matrix which associates to each row a masking vector of the samples belonging to the respective group.

        :param targets:
            The output classes.

        :return:
            The (absolute) value of the DIDI.
        """
        # onehot encode targets if this has not been done yet
        targets = probabilities.get_onehot(vector=targets, onehot_binary=True) if targets.ndim == 1 else targets
        didi, classes = 0.0, targets.shape[1]
        # compute averages per class by aggregating the (onehot encoded) targets via column, i.e., each class
        total_averages_per_class = targets.mean(axis=0)
        for protected_group in indicator_matrix:
            # subset of the targets having <label> as protected feature (i.e., the current protected group)
            protected_targets = targets[protected_group]
            if len(protected_targets) > 0:
                # list of deviations from the total percentage of samples respectively to each target class
                protected_averages_per_class = protected_targets.mean(axis=0)
                # total deviation (partial didi) respectively to each protected group
                didi += np.abs(total_averages_per_class - protected_averages_per_class).sum()
        return didi

    @staticmethod
    def regression_didi(indicator_matrix: np.ndarray, targets: np.ndarray):
        """Computes the Disparate Impact Discrimination Index for Regression Tasks.

        :param indicator_matrix:
            A matrix which associates to each row a masking vector of the samples belonging to the respective group.

        :param targets:
            The output values.

        :return:
            The (absolute) value of the DIDI.
        """
        didi = 0.0
        total_average = np.mean(targets)
        for protected_group in indicator_matrix:
            # subset of the targets having <label> as protected feature (i.e., the current protected group)
            protected_targets = targets[protected_group]
            if len(protected_targets) > 0:
                # total deviation (partial didi) respectively to each protected group
                protected_average = np.mean(protected_targets)
                didi += abs(protected_average - total_average)
        return didi

    def __init__(self, classification: bool, protected: str, percentage: bool = True, name: str = 'didi'):
        """
        :param classification:
            Whether the task is a classification (True) or a regression (False) task.

        :param protected:
            The name of the protected feature.

            During the solving process, the algorithm checks inside the data for column names starting with the given
            feature name: if a single column is found this is assumed to be a categorical column (thus, the number of
            classes is inferred from the unique values in the column), otherwise, if multiple columns are found this is
            assumed to be a one-hot encoded version of the categorical column (thus, the number of classes is inferred
            from the number of columns).

        :param percentage:
            Whether or not to normalize the DIDI index of the predictions over the DIDI index of the ground truths.

        :param name:
            The name of the metric.
        """
        super(DIDI, self).__init__(name=name)

        self.classification: bool = classification
        """Whether the task is a classification (True) or a regression (False) task."""

        self.protected: str = protected
        """The name of the protected feature."""

        self.percentage: bool = percentage
        """Whether or not to normalize the DIDI index of the predictions over the DIDI index of the ground truths."""

    def __call__(self, x: pd.DataFrame, y, p) -> float:
        m = self.get_indicator_matrix(x=x, protected=self.protected)
        if self.classification:
            p = probabilities.get_classes(p)
            didi_p = DIDI.classification_didi(indicator_matrix=m, targets=p)
            didi_y = DIDI.classification_didi(indicator_matrix=m, targets=y) if self.percentage else 1.0
        else:
            didi_p = DIDI.regression_didi(indicator_matrix=m, targets=p)
            didi_y = DIDI.regression_didi(indicator_matrix=m, targets=y) if self.percentage else 1.0
        # handle division by zero
        if didi_y == 0.0:
            return 0.0 if didi_p == 0.0 else float('inf')
        else:
            return didi_p / didi_y


class MonotonicViolation(Metric):
    """Violation of the Monotonicity Shape Constraint."""

    def __init__(self,
                 monotonicities_fn: Callable,
                 aggregation: str = 'average',
                 eps: float = 1e-3,
                 name: str = 'monotonicity'):
        """
        :param monotonicities_fn:
            Function having signature f(x) -> M where x is the input data and M is a data structure representing the
            expected monotonicities. If the computation of the monotonicities is heavy and you call the metric multiple
            times with the same input, it would be better to precompute the matrix on the input x and pass a dummy
            function that ignores the x parameter and just returns the precomputed matrix, e.g:

            .. code-block:: python

                M = monotonicities_fn(x)
                metric = MonotonicViolation(monotonicities_fn=lambda x: M)

            The data structure M can be either a NxN numpy matrix (where |x| = N) which associates to each entry (i, j)
            a value of -1, 0,  or 1 depending on the expected monotonicity between x[i] and x[j], or directly a list of
            tuples (i, j) meaning that x[i] is expected to be greater than x[j].

            E.g., if x = [0, 1, 2, 3], with 0 < 1 < 2 < 3, then either M =
                |  0, -1, -1, -1 |
                |  1,  0, -1, -1 |
                |  1,  1,  0, -1 |
                |  1,  1,  1,  0 |
            or M = [(1, 0), (2, 0), (2, 1), (3, 0), (3, 1), (3, 2)].

        :param aggregation:
            The aggregation type, which can be 'average' (i.e., it returns the average constraint violation in terms of
            output difference), 'percentage' (i.e., it returns the violation in terms of average number of violations),
            or 'feasible' (i.e., it returns a binary value depending on whether there is at least on violation).

        :param eps:
            The slack value under which a violation is considered to be acceptable.

        :param name:
            The name of the metric.
        """
        super(MonotonicViolation, self).__init__(name=name)

        self.monotonicities_fn: Callable = monotonicities_fn
        """Function having signature f(x) -> M where x is the input data and M is the monotonicities matrix."""

        self.eps: float = eps
        """The slack value under which a violation is considered to be acceptable."""

        self.aggregate: Callable
        """The aggregation function."""

        if aggregation == 'average':
            self.aggregate = lambda violations: np.mean(violations)
        elif aggregation == 'percentage':
            self.aggregate = lambda violations: np.mean(violations > 0)
        elif aggregation == 'feasible':
            self.aggregate = lambda violations: int(np.all(violations <= 0))
        else:
            raise AssertionError(f"'{aggregation}' is not a valid aggregation kind")

    def __call__(self, x, y: np.ndarray, p: np.ndarray) -> float:
        mono = self.monotonicities_fn(x)
        # handle the case in which a monotonicity matrix is passed instead of a list of monotonicities
        if isinstance(mono, np.ndarray) and mono.ndim == 2:
            mono = [(hi, li) for hi, row in enumerate(mono) for li, m in enumerate(row) if m == 1]
        # compute violations as the difference between p[lower_indices] and p[higher_indices]
        violations = p[[li for _, li in mono]] - p[[hi for hi, _ in mono]] if len(mono) > 0 else np.array([0.0])
        # filter out values under the threshold
        violations[violations < self.eps] = 0.0
        return self.aggregate(violations)


class CausalIndependence(Metric):
    """Measure of Causal Independence obtained as the weight(s) of a linear regressor trained on a sub-dataset with the
    given feature(s) only.

    Let A be the (N, K) matrix of inputs related to the given features, where <N> is the number of data samples and <K>
    is the number of given features, then A[i, j] represents the i-th value of the j-th given feature in the dataset.
    We measure the casual relationship between each given feature and the output by training a linear regressor on the
    pair (A, y), thus solving the linear system: "A.T @ w = (A.T @ A) @ y", with w being the learned weights respective
    to each one of the given features. The level of independence between A[i] and y is thus measured as abs(w[i]), as
    lower w[i] means that the i-th feature is not informative enough to let the regressor capture a trend to predict y.
    """

    def __init__(self, features: List, aggregation: Union[None, str, Callable] = 'sum', name: str = 'independence'):
        """
        :param features:
            The list of features to inspect.

        :param aggregation:
            The aggregation policy in case of multiple features. It can be either a string representing a numpy method
            for aggregation (e.g., 'sum', 'mean', 'max', 'min', 'std'), a custom callable function taking the vector
            'w' as parameter, or None to get in output the weight for each feature without any aggregation.

        :param name:
            The name of the metric.
        """
        super(CausalIndependence, self).__init__(name=name)

        self.features: List = features
        """The list of features to inspect."""

        self.aggregation: Optional[Callable] = None
        """The aggregation policy in case of multiple features."""

        if aggregation is None:
            # if the given aggregation is None, return the weights as a dictionary indexed by feature
            self.aggregation = lambda w: {f: v for f, v in zip(self.features, w)}
        elif isinstance(aggregation, str):
            # if the given aggregation is a string representing a numpy method, use np.<method_name>()
            assert hasattr(np, aggregation), f"'{aggregation}' is not a supported aggregation policy"
            self.aggregation = getattr(np, aggregation)
        else:
            self.aggregation = aggregation

    def __call__(self, x, y: np.ndarray, p: np.ndarray) -> Union[float, Dict[str, float]]:
        w, _, _, _ = np.linalg.lstsq(x[self.features], p, rcond=None)
        return self.aggregation(np.abs(w))



"""Basic Logger Interface."""
import logging
from typing import Dict, Set, Union, List, Any, Optional

from moving_targets.callbacks.callback import Callback
from moving_targets.util.errors import not_implemented_message


class Logger(Callback):
    """Basic interface for a Moving Targets Logger callback."""

    def __init__(self):
        """"""
        super(Logger, self).__init__()

        self._cache: Dict = {}
        """The internally stored cache."""

    def log(self, **cache):
        """Adds the given keyword argument to the inner cache.

        :param cache:
            Key-value pairs to be added to the cache (if the key is already present, the value is overwrote).
        """
        self._cache.update(cache)


class StatsLogger(Logger):

    @staticmethod
    def _parameters() -> Set[str]:
        """The set of all the possible parameters whose statistics may be logged."""
        raise NotImplementedError(not_implemented_message('_parameters', static=True))

    def __init__(self, stats: Union[bool, List[str]], name: Optional[str]):
        """
        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param name:
            Either a custom logger name or None (in that case, the root logger is used and no prefix is added).
        """
        super(StatsLogger, self).__init__()

        # handle stats logging ache optionally check for parameter correctness
        if isinstance(stats, bool):
            stats = self._parameters() if stats else set()
        else:
            parameters = self._parameters()
            for param in stats:
                assert param in parameters, f"Parameter '{param}' is not valid for stats logging."

        self.parameters: Set[str] = set(stats)
        """The set of parameters whose statistics must be logged."""

        self.prefix: str = '' if name is None else f'{name.lower()}/'
        """The prefix to prepend when logging a stat."""

        self.logger: Optional[logging.Logger] = logging.root if name is None else logging.getLogger(name)
        """A logger instance."""

    def _log_stats(self, **params: Any):
        """Logs the statistics about the parameters if they are included in the inner 'parameters' field.

        :param params:
            A dictionary which associates to each parameter name its respective value.
        """
        to_cache = {}
        # log data:
        #   > if the parameter is in the stats set, we cache it and log it at info level
        #   > otherwise we don't cache it and log it at debug level
        for name, value in params.items():
            if name in self.parameters:
                to_cache[f'{self.prefix}{name}'] = value
                self.logger.info(f'{name} = {value}')
            else:
                self.logger.debug(f'{name} = {value}')
        self.log(**to_cache)



"""Basic Optimizer Interface."""
from typing import Callable

import numpy as np


class Optimizer:
    """Basic interface for a Moving Targets Master Optimizer."""

    def __init__(self, base):
        """
        :param base:
            Either a fixed floating point value representing the initial value for the hyper-parameter to optimize, or
            a wrapped custom optimizer which returns a dynamic value to reduce by the given factor after each iteration.
        """

        self.base: Callable = base if isinstance(base, Optimizer) else (lambda macs, x, y, p: base)
        """The base value of the hyper-parameter to optimize or a base optimizer wrapper."""

    def __call__(self, macs, x, y: np.ndarray, p: np.ndarray) -> float:
        """Returns the current value for the hyper-parameter, computed via the custom optimization strategy.

        :param macs:
            The `MACS` instance encapsulating the master.

        :param x:
            The training samples.

        :param y:
            The training labels.

        :param p:
            The `Learner` predictions.

        :return:
            The current value of the hyper-parameter to optimize.
        """
        return self.base(macs=macs, x=x, y=y, p=p)



"""Basic Metric Interface."""
from typing import Dict, Union

import numpy as np

from moving_targets.util.errors import not_implemented_message


class Metric:
    """Basic interface for a Moving Targets Metric."""

    def __init__(self, name: str):
        """
        :param name:
            The name of the metric.
        """
        super(Metric, self).__init__()

        self.__name__: str = name
        """The name of the metric."""

    def __call__(self, x, y: np.ndarray, p: np.ndarray) -> Union[float, Dict[str, float]]:
        """Core method used to compute the metric value.

        :param x:
            The input matrix.

        :param y:
            The vector of ground truths.

        :param p:
            The vector of predictions.

        :return:
            Either a single value representing the metric, or a dictionary of multiple values for multi-metrics.
        """
        raise NotImplementedError(not_implemented_message(name='__call__'))



"""Moving Targets Backends for Masters."""

from moving_targets.masters.backends.backend import Backend
from moving_targets.masters.backends.cplex_backend import CplexBackend
from moving_targets.masters.backends.cvxpy_backend import CvxpyBackend
from moving_targets.masters.backends.gurobi_backend import GurobiBackend
from moving_targets.masters.backends.numpy_backend import NumpyBackend

_BACKEND_ALIASES: dict = {
    'cvxpy': CvxpyBackend,
    'cplex': CplexBackend,
    'gurobi': GurobiBackend,
}


def get_backend(alias: str, **backend_kwargs) -> Backend:
    backend_type = _BACKEND_ALIASES.get(alias)
    assert backend_type is not None, f"{alias} is not a valid alias for a backend"
    return backend_type(**backend_kwargs)



from typing import Union

import numpy as np

from moving_targets.masters.optimizers.optimizer import Optimizer


class ConstantSlope(Optimizer):
    """Optimizer which reduces the hyper-parameter value by the iteration number."""

    def __init__(self, base: Union[float, Optimizer]):
        """
        :param base:
            Either a fixed floating point value representing the initial value for the hyper-parameter to optimize, or
            a wrapped custom optimizer which returns a dynamic value to reduce by the given factor after each iteration.
        """
        super(ConstantSlope, self).__init__(base=base)

    def __call__(self, macs, x, y: np.ndarray, p: np.ndarray) -> float:
        value = super(ConstantSlope, self).__call__(macs, x, y, p)
        return value / macs.iteration


class ExponentialSlope(Optimizer):
    """Optimizer which reduces the hyper-parameter value by the same factor at each iteration."""

    def __init__(self, base: Union[float, Optimizer], slope: float = 2):
        """
        :param base:
            Either a fixed floating point value representing the initial value for the hyper-parameter to optimize, or
            a wrapped custom optimizer which returns a dynamic value to reduce by the given factor after each iteration.

        :param slope:
            The base of the exponential slope used to decrease the hyper-parameter value (e.g., starting from a value
            of 10 and using a slope of 2, the updates will lead to the sequence 10 -> 5 -> 2.5 -> 1.25 -> ...).
        """
        super(ExponentialSlope, self).__init__(base=base)

        self.slope: float = slope
        """The slope used to decrease the hyper-parameter value"""

    def __call__(self, macs, x, y: np.ndarray, p: np.ndarray) -> float:
        value = super(ExponentialSlope, self).__call__(macs, x, y, p)
        return value / (self.slope ** macs.iteration)



from abc import ABC
from typing import Callable

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.masters.losses.loss import WeightedLoss
from moving_targets.util import probabilities
from moving_targets.util.errors import not_implemented_message


class ClassificationLoss(WeightedLoss, ABC):
    """Basic interface for a Moving Targets Classification Loss."""

    def __init__(self, name: str):
        """
        :param name:
            The name of the loss.
        """
        super(ClassificationLoss, self).__init__(aggregation='sum_of_means', name=name)

    def _losses(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        if model_variables.ndim == 1:
            # for binary classification tasks the model variables are in the form of a 1d array, thus in this case we
            # can compute the loss by calling the strategy function two times, one per each binary value
            first_term = self._strategy(backend, 1 - numeric_variables, 1 - model_variables)
            second_term = self._strategy(backend, numeric_variables, model_variables)
            return first_term + second_term
        else:
            return self._strategy(backend, numeric_variables, model_variables)

    def _strategy(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        """Template method used as the loss strategy inside the _losses() method.

        This is useful since the majority of classification losses deal with probabilities and, therefore, with 2D
        variables that cannot be handled properly by the backends primitives which expect a 1D array. In order to deal
        with this, the default implementation of the _losses() abstract method is in charge of flattening the input
        data, calling the strategy on the flattened vectors and, finally, reshaping the variables in order to be
        consistent for the multiplication with the sample weights. When it comes to classification losses that use
        class values instead of probabilities (e.g., hamming distance), however, this cannot be done, thus an explicit
        _losses() implementation must be passed.
        """
        raise NotImplementedError(not_implemented_message(name='_strategy'))


class HammingDistance(ClassificationLoss):
    """Hamming Distance."""

    def __init__(self, labelling: bool = False, name: str = 'hamming_distance'):
        """
        :param labelling:
            The kind of classification task, either 'classification', 'labelling' or 'auto' for automatic task
            inference depending on the given numeric variables.

        :param name:
            Whether the master represents a labelling or a classification task.
        """
        super(HammingDistance, self).__init__(name=name)

        self.labelling: bool = labelling
        """Whether the master represents a labelling or a classification task."""

    def _strategy(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        # it should be 1 - sum([mv[i] * nv[i] for i in range(num_classes)]), but this creates problems due to the
        # "backend.sum()" call, therefore the defined strategy simply returns raw the losses which are then handled
        # in the _losses() function
        return model_variables * numeric_variables

    def _losses(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        # the hamming distance is computed on class values, not on probabilities, thus if we do not get integer values
        # we first obtain classes/labels then we bring them back to a onehot encoded version for compatibility (this is
        # necessary only for the multiclass task, since classes are given in a 1d vector but model variables are 2d)
        if not np.issubdtype(numeric_variables.dtype, np.integer):
            numeric_variables = probabilities.get_classes(numeric_variables, labelling=self.labelling)
            if numeric_variables.ndim == 1 and model_variables.ndim == 2:
                numeric_variables = probabilities.get_onehot(vector=numeric_variables,
                                                             classes=model_variables.shape[1],
                                                             onehot_binary=True)
        losses = super(HammingDistance, self)._losses(backend, numeric_variables, model_variables)
        # as said before, the "1 - ..." part is moved here due to computational problems, in particular, for binary
        # tasks we just need to do (1 - losses) while for multiclass tasks it must be (1 - sum(losses, axis=1))
        return 1 - (losses if losses.ndim == 1 else np.array([backend.sum(row) for row in losses]))


class CrossEntropy(ClassificationLoss):
    """Negative Log-Likelihood Loss."""

    def __init__(self, clip_value: float = 1e-15, name: str = 'crossentropy'):
        """
        :param clip_value:
            The clipping value to be used to avoid numerical errors with the log.

        :param name:
            The name of the metric.
        """
        super(CrossEntropy, self).__init__(name=name)

        self.clip_value: float = clip_value
        """The clipping value to be used to avoid numerical errors with the log."""

    def _strategy(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        numeric_variables = numeric_variables.clip(min=self.clip_value, max=1 - self.clip_value)
        return -model_variables * np.log(numeric_variables)


class ReversedCrossEntropy(ClassificationLoss):
    """Reversed Negative Log-Likelihood Loss."""

    def __init__(self, binary: bool = False, clip_value: float = 1e-15, name: str = 'crossentropy'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            If the model variables are binary, their value will be clipped, otherwise for continuous variables there
            will be no explicit clipping in order to let the backend take care of choosing the most ideal values.

        :param clip_value:
            The clipping value to be used to avoid numerical errors with the log.

            This will be used only for binary model variables.

        :param name:
            The name of the metric.
        """
        if binary:
            # if we expect binary variables, we clip them into [<clip_value>, 1 - <clip_value>] relying on the formula:
            #   (1 - 2 * <clip_value>) * <m> + <clip_value>
            # since this formula will evaluate to:
            #   > (1 - 2 * <clip_value>) * 0 + <clip_value> = <clip_value>, if <m> = 0
            #   > (1 - 2 * <clip_value>) * 1 + <clip_value> = 1 - <clip_value>, if <m> = 1
            clip = lambda m: (1.0 - 2 * clip_value) * m + clip_value
        else:
            # otherwise we simply return the continuous variables
            clip = lambda m: m

        super(ReversedCrossEntropy, self).__init__(name=name)

        self._clip: Callable = clip
        """The clipping function to be called before the logarithm."""

    def _strategy(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        clipped_variables = self._clip(model_variables)
        return -numeric_variables * backend.log(clipped_variables)


class SymmetricCrossEntropy(ClassificationLoss):
    """Reversed Negative Log-Likelihood Loss."""

    def __init__(self, binary: bool = False, clip_value: float = 1e-15, name: str = 'crossentropy'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            If the model variables are binary, their value will be clipped, otherwise for continuous variables there
            will be no explicit clipping in order to let the backend take care of choosing the most ideal values.

        :param clip_value:
            The clipping value to be used to avoid numerical errors with the log.

        :param name:
            The name of the metric.
        """

        if binary:
            # if we expect binary variables, we clip them into [<clip_value>, 1 - <clip_value>] relying on the formula:
            #   (1 - 2 * <clip_value>) * <m> + <clip_value>
            # since this formula will evaluate to:
            #   > (1 - 2 * <clip_value>) * 0 + <clip_value> = <clip_value>, if <m> = 0
            #   > (1 - 2 * <clip_value>) * 1 + <clip_value> = 1 - <clip_value>, if <m> = 1
            clip = lambda m: (1.0 - 2 * clip_value) * m + clip_value
        else:
            # otherwise we simply return the continuous variables
            clip = lambda m: m

        super(SymmetricCrossEntropy, self).__init__(name=name)

        self.clip_value: float = clip_value
        """The clipping value to be used to avoid numerical errors with the log."""

        self._clip: Callable = clip
        """The clipping function to be called before the logarithm."""

    def _strategy(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        clipped_variables = self._clip(model_variables)
        standard_ce = -model_variables * np.log(numeric_variables.clip(min=self.clip_value, max=1 - self.clip_value))
        reversed_ce = -numeric_variables * backend.log(clipped_variables)
        return standard_ce + reversed_ce



"""Basic Backend Interface."""
import logging
from typing import Any, Union, List, Optional

import numpy as np

from moving_targets.util.errors import not_implemented_message, BackendError


class Backend:
    """Basic Interface for a Moving Targets Master Backend."""

    _LOGGER: logging.Logger = logging.getLogger('Backend')

    def _nested_names(self, *keys: int, name: Optional[str]) -> List:
        """Generates an keys-dimensional list of nested lists containing the correct variables names when an array of
        variables is wanted instead of a single one by appending the variable position in the keys-dimensional tensor.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param name:
            The common name for the variables.

        :return:
            The keys-dimensional structure of nested lists.
        """
        if len(keys) == 1:
            return [None if name is None else f'{name}_{i}' for i in range(keys[0])]
        else:
            return [self._nested_names(*keys[1:], name=None if name is None else f'{name}_{i}') for i in range(keys[0])]

    @classmethod
    def _aux_warning(cls, exp: Optional[str], aux: Optional[str], msg: str):
        """Logs a warning on the Backend Logger when the 'aux' parameter is not used correctly.

        :param exp:
            The expected value of the aux parameter.

        :param aux:
            The actual value of the aux parameter.

        :param msg:
            The warning explanation.
        """
        if not (exp is None and aux is None) and exp != aux:
            cls._LOGGER.warning(f"'aux={aux}' has no effect since the solver needs {exp} auxiliary variables to {msg}.")

    def __init__(self):
        super(Backend, self).__init__()

        self.model: Optional = None
        """The inner model instance."""

        self.solution: Optional = None
        """The inner solution instance."""

    def _build_model(self) -> Any:
        """Initializes a model instance.

        :return:
            The model instance.
        """
        raise NotImplementedError(not_implemented_message(name='_build_model'))

    def _solve_model(self) -> Optional:
        """Solves the optimization problem.

        :return:
            The solution instance or None if no solution is found.
        """
        raise NotImplementedError(not_implemented_message(name='_solve_model'))

    def build(self) -> Any:
        """Initializes a model instance.

        :return:
            The backend itself.
        """
        if self.model is not None:
            self._LOGGER.warning("A model instance was already present and has been overwritten. If that was " +
                                 "intentional consider calling 'backend.clear()' before creating a new one.")
        self.model = self._build_model()
        self.solution = None
        return self

    def solve(self) -> Any:
        """Solves the optimization problem.

        :return:
            The backend itself.
        """
        self.solution = self._solve_model()
        return self

    def clear(self) -> Any:
        """Clears the backend resources.

        :return:
            The backend itself.
        """
        self.model = None
        self.solution = None
        return self

    def minimize(self, cost) -> Any:
        """Sets a cost function to minimize before solving.

        :param cost:
            The cost function.

        :return:
            The backend itself.
        """
        raise NotImplementedError(not_implemented_message(name='minimize'))

    def maximize(self, cost) -> Any:
        """Sets a cost function to maximize before solving.

        :param cost:
            The cost function.

        :return:
            The backend itself.
        """
        raise NotImplementedError(not_implemented_message(name='maximize'))

    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
        """Adds a list of constraints to the model.

        :param constraints:
            The list of constraints.

        :param name:
            The constraints name.

        :return:
            The backend itself.
        """
        raise NotImplementedError(not_implemented_message(name='add_constraints'))

    def add_constraint(self, constraint, name: Optional[str] = None) -> Any:
        """Adds a constraint to the model.

        :param constraint:
            The constraint.

        :param name:
            The constraint name.

        :return:
            The backend itself.
        """
        return self.add_constraints(constraints=[constraint], name=name)

    def add_binary_variables(self, *keys: int, name: Optional[str] = None) -> np.ndarray:
        """Creates an array of binary model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param name:
            The variables name.

        :return:
            The array of binary model variables.
        """
        return self.add_variables(*keys, vtype='binary', name=name, lb=0, ub=1)

    def add_binary_variable(self, name: Optional[str] = None) -> Any:
        """Creates a binary model variable.

        :param name:
            The variable name.

        :return:
            The binary model variable.
        """
        return self.add_variable(vtype='binary', name=name, lb=0, ub=1)

    def add_integer_variables(self,
                              *keys: int,
                              lb: float = 0,
                              ub: float = float('inf'),
                              name: Optional[str] = None) -> np.ndarray:
        """Creates an array of integer model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param lb:
            The variables lower bound.

        :param ub:
            The variables upper bound.

        :param name:
            The variables name.

        :return:
            The array of integer model variables.
        """
        return self.add_variables(*keys, vtype='integer', name=name, lb=lb, ub=ub)

    def add_integer_variable(self,
                             lb: float = 0,
                             ub: float = float('inf'),
                             name: Optional[str] = None) -> np.ndarray:
        """Creates an integer model variable.

        :param lb:
            The variable lower bound.

        :param ub:
            The variable upper bound.

        :param name:
            The variable name.

        :return:
            The integer model variable.
        """
        return self.add_variable(vtype='integer', name=name, lb=lb, ub=ub)

    def add_continuous_variables(self,
                                 *keys: int,
                                 lb: float = -float('inf'),
                                 ub: float = float('inf'),
                                 name: Optional[str] = None) -> np.ndarray:
        """Creates an array of continuous model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param lb:
            The variables lower bound.

        :param ub:
            The variables upper bound.

        :param name:
            The variables name.

        :return:
            The array of continuous model variables.
        """
        return self.add_variables(*keys, vtype='continuous', name=name, lb=lb, ub=ub)

    def add_continuous_variable(self,
                                lb: float = -float('inf'),
                                ub: float = float('inf'),
                                name: Optional[str] = None) -> np.ndarray:
        """Creates a continuous model variable.

        :param lb:
            The variable lower bound.

        :param ub:
            The variable upper bound.

        :param name:
            The variable name.

        :return:
            The continuous model variable.
        """
        return self.add_variable(vtype='continuous', name=name, lb=lb, ub=ub)

    def add_constant(self, val: Any, vtype: str = 'continuous', name: Optional[str] = None) -> Any:
        """Creates a model variable with a constant value.

        :param val:
            The numerical value of the constant.

        :param vtype:
            The constant type, by default this is 'continuous'.

        :param name:
            The constant name.

        :return:
            A model variable with value fixed to the given one via a constraint.
        """
        return self.add_variable(vtype=vtype, lb=val, ub=val, name=name)

    def add_constants(self, val: np.ndarray, vtype: str = 'continuous', name: Optional[str] = None) -> np.ndarray:
        """Creates an array of model variables with constant values.

        :param val:
            The array of numerical values of the constants.

        :param vtype:
            The constants type, by default this is 'continuous'.

        :param name:
            The constants name.

        :return:
            An array of model variables with value fixed to the given ones via a set of constraints.
        """
        # a default strategy to create an array of constants is to compute their names leveraging the utility function,
        # creating a mono-dimensional list of variables with correct names and fixed lower/upper bounds to the given
        # values then, eventually, reshaping then into the correct shape
        names = np.array(self._nested_names(*val.shape, name=name)).flatten()
        var = [self.add_variable(vtype=vtype, lb=v, ub=v, name=n) for v, n in zip(val.flatten(), names)]
        return np.reshape(var, val.shape)

    def add_variable(self, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> Any:
        """Creates a model variable.

        :param vtype:
            The variables type, usually 'binary' or 'continuous'.

        :param lb:
            The variable lower bound.

        :param ub:
            The variable upper bound.

        :param name:
            The variable name.

        :return:
            The model variable.
        """
        raise NotImplementedError(not_implemented_message(name='add_variable'))

    def add_variables(self, *keys: int, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> np.ndarray:
        """Creates an array of model variables.

        :param keys:
            The number of keys per dimension of the array of variables.

        :param vtype:
            The variables type, usually 'binary', 'integer', or 'continuous'.

        :param lb:
            The variables lower bound.

        :param ub:
            The variables upper bound.

        :param name:
            The variables name.

        :return:
            The array of model variables.
        """
        # a default strategy to create an array of variables is to compute their names leveraging the utility function
        # then creating a mono-dimensional list of variables and reshaping then into the correct shape
        names = np.array(self._nested_names(*keys, name=name)).flatten()
        return np.reshape([self.add_variable(vtype=vtype, lb=lb, ub=ub, name=n) for n in names], keys)

    def get_objective(self) -> float:
        """Gets the objective value of the solved model.

        :return:
            The objective value.
        """
        raise NotImplementedError(not_implemented_message(name='get_objective'))

    def get_values(self, expressions: np.ndarray) -> np.ndarray:
        """Gets the values of a vector of expressions as found in the model solution.

        :param expressions:
            The vector of expressions.

        :return:
            The vector of solution values.
        """
        raise NotImplementedError(not_implemented_message(name='get_values'))

    def get_value(self, expression: Any) -> float:
        """Gets the value of an expressions as found in the model solution.

        :param expression:
            The expressions.

        :return:
            The solution value.
        """
        return self.get_values(np.array([expression]))[0]

    def aux(self,
            expressions: Any,
            aux_vtype: Optional[str] = 'continuous',
            aux_lb: float = -float('inf'),
            aux_ub: float = float('inf'),
            aux_name: Optional[str] = None) -> Any:
        """If the 'aux_vtype' parameter is None, it simply return the input expressions, otherwise it builds and return
        an auxiliary variable (or a vector of variables) having the given vtype and the respective other properties,
        which is equal to the given expression (or vector of expressions). Using auxiliary variables may come in handy
        when dealing with huge datasets since they can considerably speedup the model formulation; still, imposing
        equality constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :param expressions:
            Either a single expression or a vector of expressions.

        :param aux_vtype:
            Either None (or 'auto') or the auxiliary variables type, usually 'binary', 'integer', or 'continuous'.

        :param aux_lb:
            The auxiliary variables lower bound.

        :param aux_ub:
            The auxiliary variables upper bound.

        :param aux_name:
            The auxiliary variables name.

        :return:
            Either a single auxiliary variable or the array of auxiliary variables.
        """
        if aux_vtype is None or aux_vtype == 'auto':
            return expressions
        else:
            if isinstance(expressions, np.ndarray):
                expressions, is_numpy = expressions, True
            else:
                expressions, is_numpy = np.array([expressions]), False
            variables = self.add_variables(*expressions.shape, vtype=aux_vtype, lb=aux_lb, ub=aux_ub, name=aux_name)
            self.add_constraints([v == e for v, e in zip(variables.flatten(), expressions.flatten())])
            return variables if is_numpy else variables[0]

    def sum(self, a: np.ndarray, aux: Optional[str] = None) -> Any:
        """Computes the sum of an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The variables sum.
        """
        raise NotImplementedError(not_implemented_message(name='sum'))

    def square(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        """Computes the squared values over an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The vector of squared values.

        :raise `BackendError`:
            If the backend cannot handle squared values.
        """
        raise BackendError(unsupported='squared values')

    def sqrt(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        """Computes the squared roots over an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The vector of squared roots.
        """
        raise BackendError(unsupported='squared roots')

    def abs(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        """Computes the absolute values over a vector of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The vector of absolute values.

        :raise `BackendError`:
            If the backend cannot handle absolute values.
        """
        raise BackendError(unsupported='absolute values')

    def log(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        """Computes the logarithms over an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The vector of logarithms.

        :raise `BackendError`:
            If the backend cannot handle logarithms.
        """
        raise BackendError(unsupported='logarithms')

    def mean(self, a: np.ndarray, aux: Optional[str] = None) -> Any:
        """Computes the mean of an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The variables mean.
        """
        sum_expression = self.sum(a, aux=aux)
        return self.aux(sum_expression / a.size, aux_vtype=aux)

    def var(self, a: np.ndarray, aux: Optional[str] = 'auto') -> Any:
        """Computes the variance of an array of variables.

        :param a:
            An array of model variables.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The variables variance.
        """
        mean_expression = self.mean(a, aux='continuous' if aux == 'auto' else aux)
        diff_expressions = np.reshape([ai - mean_expression for ai in a.flatten()], a.shape)
        diff_expressions = self.aux(expressions=diff_expressions, aux_vtype=aux)
        squared_expressions = self.square(diff_expressions, aux=None if aux == 'auto' else aux)
        return self.mean(squared_expressions, aux=aux)

    def add(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
        """Performs the pairwise sum between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise sums between a and b.
        """
        return self.aux(expressions=a + b, aux_vtype=aux)

    def subtract(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
        """Performs the pairwise subtraction between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise differences between a and b.
        """
        return self.aux(expressions=a - b, aux_vtype=aux)

    def multiply(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
        """Performs the pairwise product between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise products between a and b.
        """
        return self.aux(expressions=a * b, aux_vtype=aux)

    def divide(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
        """Performs the pairwise division between two arrays.

        :param a:
            The first array.

        :param b:
            The second array.

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The array of pairwise divisions between a and b.
        """
        return self.aux(expressions=a / b, aux_vtype=aux)

    def dot(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None) -> Union[Any, np.ndarray]:
        """Performs the vector product between two arrays.

        :param a:
            The first array (at most 2d).

        :param b:
            The second array (at most 2d).

        :param aux:
            The vtype of the auxiliary variables which may be added the represent the results values and, optionally,
            the partial results obtained in the computation. If 'auto' is passed, it automatically decides how to deal
            with auxiliary variables in order to maximize the computational gain without introducing formulation issues;
            if None is passed, it returns the result in the form of an expression, otherwise it builds an auxiliary
            variable bounded to the expression value. Using auxiliary variables may come in handy when dealing with
            huge datasets since they can considerably speedup the model formulation; still, imposing equality
            constraints on certain expressions may lead to solving errors due to broken model assumptions.

        :return:
            The dot product between a and b.
        """
        a = a if a.ndim == 2 else a.reshape((1, len(a)))
        b = b if b.ndim == 2 else b.reshape((len(b), 1))
        # we use a list of lists instead of an array to let numpy handle the data type on its own
        r = []
        for i, row in enumerate(a):
            r.append([])
            for j, column in enumerate(b.T):
                pairwise_products = self.multiply(a=row, b=column, aux=aux)
                r[i].append(self.sum(pairwise_products, aux=aux))
        r = np.array(r)
        return r[0, 0] if r.size == 1 else r.squeeze()



"""Core of the Moving Targets algorithm."""
import time
import warnings
from typing import List, Dict, Callable, Union, Optional, Any, Set

import numpy as np

from moving_targets.callbacks.callback import Callback
from moving_targets.callbacks.console_logger import ConsoleLogger
from moving_targets.callbacks.file_logger import FileLogger
from moving_targets.callbacks.history import History
from moving_targets.callbacks.logger import Logger, StatsLogger
from moving_targets.learners.learner import Learner
from moving_targets.masters.master import Master
from moving_targets.metrics.metric import Metric
from moving_targets.util.typing import Dataset, is_numeric


class MACS(StatsLogger):
    """Model-Agnostic Constraint Satisfaction.

    This class contains the core algorithm of Moving Targets. It leverages the `Learner` and `Master` instances to
    iteratively refine the predictions during the training phase and, eventually, it evaluates the solutions based on
    the given list of `Metric` objects.
    """

    @staticmethod
    def _parameters() -> Set[str]:
        return {'iteration', 'elapsed_time'}

    def __init__(self,
                 master: Master,
                 learner: Learner,
                 init_step: str = 'pretraining',
                 metrics: List[Metric] = (),
                 stats: Union[bool, List[str]] = False):
        """
        :param master:
            A `Master` instance.

        :param learner:
            A `Learner` instance.

        :param init_step:
            The initial step of the algorithm, which can be either 'pretraining' or 'projection'.

        :param metrics:
            A list of `Metric` instances to evaluate the final solution.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in
            ['iteration', 'elapsed_time'] whose statistics must be logged.
        """
        super(MACS, self).__init__(stats=stats, name='MACS')
        assert init_step in ['pretraining', 'projection'], f"'{init_step}' is not a valid initial step"

        self.master: Master = master
        """The `Master` instance."""

        self.learner: Learner = learner
        """The `Learner` instance."""

        self.init_step: str = init_step
        """The initial step of the algorithm, which can be either 'pretraining' or 'projection'."""

        self.iteration: Optional[int] = None
        """The current `MACS` iteration."""

        self.fitted: bool = False
        """Whether or not the `Learner` has been fitted at least once."""

        self.metrics: List[Metric] = list(metrics)
        """A list of `Metric` instances to evaluate the final solution."""

        self._history: History = History()
        """The internal `History` object which is returned at the end of the training."""

        self._time: Optional[float] = None
        """An auxiliary variable to keep track of the elapsed time between iterations."""

    def fit(self,
            x,
            y: np.ndarray,
            iterations: int = 1,
            sample_weight: Optional[np.ndarray] = None,
            val_data: Optional[Dataset] = None,
            callbacks: List[Callback] = (),
            verbose: Union[int, bool] = 2) -> History:
        """Fits the `Learner` based on the Moving Targets iterative procedure.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param iterations:
            The number of algorithm iterations.

        :param sample_weight:
            The (optional) array of sample weights.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

            .. code-block:: python

                val_data = dict(train=(xtr, ytr), validation=(xvl, yvl), test=(xts, yts))

        :param callbacks:
            A list of `Callback` instances.

        :param verbose:
            Either a boolean or an int representing the verbosity value, such that:

            - `0` or `False` create no logger;
            - `1` creates a simple console logger with elapsed time only;
            - `2` or `True` create a more complete console logger with cached data at the end of each iterations.

        :return:
            An instance of the `History` object containing the training history.

        :raise `AssertionError`:
            If the number of iteration is negative, or is zero and the initial step is 'pretraining'.
        """
        # check user input
        assert iterations >= 0, f'the number of iterations should be non-negative, but it is {iterations}'
        assert self.init_step == 'pretraining' or iterations > 0, 'if projection, iterations cannot be zero'
        assert verbose in [True, False, 0, 1, 2], 'unknown verbosity'
        val_data = {} if val_data is None else (val_data if isinstance(val_data, dict) else {'val': val_data})

        # handle callbacks and verbosity (check for type instance since 1 == True and vice versa)
        callbacks = list(callbacks)
        if verbose == 1 and not isinstance(verbose, bool):
            callbacks += [ConsoleLogger()]
        elif verbose == 2 or verbose is True:
            callbacks += [FileLogger(routines=['on_pretraining_end', 'on_iteration_end'])]
        self._update_callbacks(callbacks, lambda c: c.on_process_start(macs=self, x=x, y=y, val_data=val_data))

        # handle pretraining
        if self.init_step == 'pretraining':
            self.iteration = 0
            self._update_callbacks(callbacks, lambda c: c.on_pretraining_start(macs=self, x=x, y=y, val_data=val_data))
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self.learner.fit(x=x, y=y, sample_weight=sample_weight)
            self.fitted = True
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self._update_callbacks(callbacks, lambda c: c.on_pretraining_end(macs=self, x=x, y=y, val_data=val_data))

        # algorithm core
        for self.iteration in range(1, iterations + 1):
            self._update_callbacks(callbacks, lambda c: c.on_iteration_start(macs=self, x=x, y=y, val_data=val_data))
            self._update_callbacks(callbacks, lambda c: c.on_adjustment_start(macs=self, x=x, y=y, val_data=val_data))
            # ---------------------------------------------- MASTER  STEP ----------------------------------------------
            adjusted_y = self.master.adjust_targets(x=x, y=y)
            if adjusted_y is None:
                # in case of no valid solution, raise a warning and stop the training loop
                warnings.warn(f'No solution found at iteration {self.iteration}, stop training. ' +
                              f'This may have been cause either by time limit or by model infeasibility.')
                break
            # ---------------------------------------------- MASTER  STEP ----------------------------------------------
            self._update_callbacks(callbacks, lambda c: c.on_adjustment_end(macs=self, x=x, y=y, val_data=val_data,
                                                                            adjusted_y=adjusted_y))
            self._update_callbacks(callbacks, lambda c: c.on_training_start(macs=self, x=x, y=y, val_data=val_data))
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self.learner.fit(x=x, y=adjusted_y, sample_weight=sample_weight)
            self.fitted = True
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self._update_callbacks(callbacks, lambda c: c.on_training_end(macs=self, x=x, y=y, val_data=val_data))
            self._update_callbacks(callbacks, lambda c: c.on_iteration_end(macs=self, x=x, y=y, val_data=val_data))
        self._update_callbacks(callbacks, lambda c: c.on_process_end(macs=self, val_data=val_data))
        return self._history

    def predict(self, x) -> Any:
        """Uses the previously trained `Learner` to predict labels from input samples.

        :param x:
            The matrix/dataframe of input samples.

        :return:
            The vector of predicted labels.

        :raise `AssertionError`:
            If the `Learner` has not been fitted yet.
        """
        assert self.fitted, 'The model has not been fitted yet, please call method .fit()'
        return self.learner.predict(x)

    def evaluate(self, x, y) -> Dict[str, float]:
        """Evaluates the performances of the model based on the given set of metrics.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :return:
            The dictionary of evaluated metrics.

        :raise `AssertionError`:
            If the `Learner` has not been fitted yet.
        """
        return self._compute_metrics(x=x, y=y, p=self.predict(x), metrics=self.metrics, prefix=None)

    def on_iteration_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._log_stats(iteration=self.iteration)
        self._time = time.time()

    def on_adjustment_end(self, macs, x, y: np.ndarray, adjusted_y: np.ndarray, val_data: Optional[Dataset]):
        # log metrics on adjusted data
        self.log(**self._compute_metrics(x=x, y=y, p=adjusted_y, metrics=self.metrics, prefix='adjusted'))

    def on_training_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # log metrics on training data
        self.log(**self._compute_metrics(x=x, y=y, p=self.predict(x), metrics=self.metrics, prefix='predictions'))

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # log metrics on validation data
        for name, (xv, yv) in val_data.items():
            self.log(**self._compute_metrics(x=xv, y=yv, p=self.predict(xv), metrics=self.metrics, prefix=name))
        # log stats
        self._log_stats(elapsed_time=time.time() - self._time)

    def _compute_metrics(self,
                         x,
                         y: np.ndarray,
                         p: np.ndarray,
                         metrics: List[Metric],
                         prefix: Optional[str] = None) -> Dict[str, float]:
        """Compute the metrics results on the given data.

        :param x:
            The input data.

        :param y:
            The output targets.

        :param p:
            The predicted targets.

        :param metrics:
            The list of `Metric` instances for the evaluation.

        :param prefix:
            A prefix to be prepended to the metric name in the results dictionary.

        :return:
            A dictionary of metric results indexed by metric name (optionally with a prefix).
        """
        prefix = "" if prefix is None else f'{prefix}/'
        results = {}
        for metric in metrics:
            value = metric(x, y, p)
            if is_numeric(value):
                # if the metric is a single metric, directly assign the value
                results[f'{prefix}{metric.__name__}'] = value
            elif isinstance(value, dict):
                # if the metric is a multi metric, assign each value per feature by using an underscore as a separator
                for k, v in value.items():
                    results[f'{prefix}{metric.__name__}_{k}'] = v
            else:
                # handle array-like return type by giving a warning
                self.logger.warning(f"Metric '{metric.__name__}' should return either a float or a dictionary.")
                for i, v in enumerate(value):
                    results[f'{prefix}{metric.__name__}_{i}'] = v
        return results

    def _update_callbacks(self, callbacks: List[Callback], routine: Callable):
        """Runs the given routine for each one of the given callbacks, plus the routine for the `MACS` object itself
        (which is run at the beginning) and the inner `History` callback (which is run at the end).

        :param callbacks:
            The list of callbacks specified during the `fit()` call.

        :param routine:
            The callback routine (e.g., <callback>.on_iteration_start()).
        """
        # run the callback routine for the MACS object itself plus the master and learner instances, for which we do
        # not need to call the 'log' method even though they are 'Logger' instances since their cache is already shared
        routine(self)
        routine(self.master)
        routine(self.learner)
        # run the callback routine for each external callback plus the history logger
        for callback in callbacks + [self._history]:
            # if the callback is a logger, share the internal cache via the 'log' method before calling the routine
            if isinstance(callback, Logger):
                callback.log(**self._cache)
            routine(callback)
        # eventually clear the internal cache
        self._cache = {}



from typing import Any, Union, List, Optional, Dict

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util.errors import MissingDependencyError, BackendError


class CvxpyBackend(Backend):
    """`Backend` implementation for the Cvxpy Solver."""

    _VTYPES: Dict[str, Any] = {
        'binary': dict(boolean=True),
        'integer': dict(integer=True),
        'continuous': dict()
    }
    """Accepted vtypes with mapping."""

    def __init__(self, solver: Optional[str] = None, **solver_args):
        """
        :param solver:
            The name of the solver (e.g., SCS, ...).

        :param solver_args:
            Parameters of the solver to be passed to the `model.solve()` function.
        """
        super(CvxpyBackend, self).__init__()

        try:
            import cvxpy
            self._cp = cvxpy
            """The lazily imported cvxpy instance."""
        except ModuleNotFoundError:
            raise MissingDependencyError(package='cvxpy')

        self.solver: Optional[str] = solver
        """The name of the solver."""

        self.solver_args: Dict[str, Any] = solver_args
        """Parameters of the solver to be passed to the `model.solve()` function."""

        self._objective: Optional = None
        """The model objective (by default, this is set to no objective by passing None)."""

    def _build_model(self) -> Any:
        return []

    def _solve_model(self) -> Optional:
        # noinspection PyTypeChecker
        model = self._cp.Problem(self._cp.Minimize(0.0) if self._objective is None else self._objective, self.model)
        try:
            model.solve(solver=self.solver, **self.solver_args)
            return None if model.status in ['infeasible', 'unbounded'] else model
        except self._cp.error.SolverError:
            raise BackendError(unsupported='the given operation')

    def minimize(self, cost) -> Any:
        self._objective = self._cp.Minimize(cost)
        return self

    def maximize(self, cost) -> Any:
        self._objective = self._cp.Maximize(cost)
        return self

    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
        self.model += constraints
        return self

    def add_variable(self, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> Any:
        if vtype not in self._VTYPES.keys():
            raise BackendError(unsupported=f"vtype '{vtype}'")
        var = self._cp.Variable(1, name=name, **self._VTYPES[vtype])
        self.add_constraints([var[0] >= lb, var[0] <= ub])
        return var

    def add_variables(self, *keys: int, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> np.ndarray:
        def _recursive_addition(_keys: List[int], _name: Optional[str]):
            key, name_fn = _keys.pop(0), lambda i: None if _name is None else f'{_name}_{i}'
            if len(_keys) == 0:
                return [self._cp.Variable((1,), name=name_fn(i), **kw) for i in range(key)]
            return [_recursive_addition(_keys=_keys.copy(), _name=name_fn(i)) for i in range(key)]

        if vtype not in self._VTYPES.keys():
            raise BackendError(unsupported=f"vtype '{vtype}'")
        kw = self._VTYPES[vtype]
        var = np.array(_recursive_addition(_keys=list(keys), _name=name))
        self.add_constraints([v[0] >= lb for v in var.flatten()] + [v[0] <= ub for v in var.flatten()])
        return var

    def get_objective(self) -> float:
        return self.solution.value

    def get_values(self, expressions: np.ndarray) -> np.ndarray:
        return np.reshape([v.value.squeeze() for v in expressions.flatten()], expressions.shape)

    def sum(self, a: np.ndarray, aux: Optional[str] = None) -> Any:
        return self.aux(expressions=a.sum(), aux_vtype=aux)

    def square(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        return self.aux(expressions=a ** 2, aux_vtype=aux)

    def abs(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        expressions = np.reshape([self._cp.abs(v) for v in a.flatten()], a.shape)
        return self.aux(expressions=expressions, aux_vtype=aux)

    def log(self, a: np.ndarray, aux: Optional[str] = None) -> np.ndarray:
        expressions = np.reshape([self._cp.log(v) for v in a.flatten()], a.shape)
        return self.aux(expressions=expressions, aux_vtype=aux)

    def var(self, a: np.ndarray, aux: Optional[str] = 'auto') -> Any:
        raise BackendError(unsupported='variance due to numerical instability')

    def multiply(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = None):
        expressions = np.reshape([self._cp.multiply(ai, bi) for ai, bi in zip(a.flatten(), b.flatten())], a.shape)
        return self.aux(expressions=expressions, aux_vtype=aux)



"""Interfaces and classes for Moving Targets Masters."""

from moving_targets.masters.classification_master import ClassificationMaster
from moving_targets.masters.master import Master
from moving_targets.masters.regression_master import RegressionMaster



"""Basic Master Interface."""
import time
from typing import Optional, Set
from typing import Union, List

import numpy as np

from moving_targets.callbacks import StatsLogger
from moving_targets.masters.backends import Backend
from moving_targets.masters.losses import Loss
from moving_targets.masters.optimizers import Optimizer
from moving_targets.util.errors import not_implemented_message
from moving_targets.util.typing import Dataset


class Master(StatsLogger):
    """Basic interface for a Moving Targets Master."""

    @staticmethod
    def _parameters() -> Set[str]:
        return {'alpha', 'beta', 'use_beta', 'y_loss', 'p_loss', 'objective', 'elapsed_time'}

    def __init__(self,
                 backend: Backend,
                 y_loss: Loss,
                 p_loss: Loss,
                 alpha: Optional[Optimizer],
                 beta: Optional[Optimizer],
                 stats: Union[bool, List[str]]):
        """
        :param backend:
            The `Backend` instance encapsulating the optimization solver.

        :param y_loss:
            A `Loss` instance used to compute the y_loss.

        :param p_loss:
            A `Loss` instance used to compute the p_loss.

        :param alpha:
            Either an optimizer for the alpha value, or None if no alpha step is wanted.

        :param beta:
            Either an optimizer for the beta value, or None if no beta step is wanted.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in ['alpha',
            'beta', 'use_beta', 'y_loss', 'p_loss', 'objective'] whose statistics must be logged.
        """

        super(Master, self).__init__(stats=stats, name='Master')

        self.backend: Backend = backend
        """The `Backend` instance encapsulating the optimization solver."""

        self.y_loss: Loss = y_loss
        """The `Loss` instance used to compute the y_loss."""

        self.p_loss: Loss = p_loss
        """The `Loss` instance used to compute the p_loss."""

        self.alpha: Optional[Optimizer] = alpha
        """The alpha optimizer, or None if no alpha step is wanted."""

        self.beta: Optional[Optimizer] = beta
        """The beta optimizer, or None if no beta step is wanted."""

        self._macs: Optional = None
        """Reference to the MACS object encapsulating the `Master`."""

        self._time: Optional[float] = None
        """An auxiliary variable to keep track of the elapsed time between iterations."""

    def log(self, **cache):
        self._macs.log(**cache)

    def on_adjustment_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._macs = macs
        self._time = time.time()

    def on_adjustment_end(self, macs, x, y: np.ndarray, adjusted_y: np.ndarray, val_data: Optional[Dataset]):
        self._log_stats(elapsed_time=time.time() - self._time)
        self._time = None
        self._macs = None

    def build(self, x, y: np.ndarray) -> np.ndarray:
        """Creates the model variables and adds the problem constraints.

        :param x:
            The training samples.

        :param y:
            The training labels.

        :return:
            The model variables.
        """
        raise NotImplementedError(not_implemented_message(name='build'))

    def use_beta(self, x, y: np.ndarray, p: np.ndarray) -> bool:
        """Decides whether or not to use the beta step for the given iteration.

        :param x:
            The training samples.

        :param y:
            The training labels.

        :param p:
            The `Learner` predictions.

        :return:
            A boolean value representing whether or not to use the beta step for the given iteration.
        """
        raise ValueError("This master has not implemented any strategy to choose between the alpha and beta steps. " +
                         "Please override a custom 'use_beta' method or set either alpha or beta to None.")

    def adjust_targets(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
        """Core function of the `Master` object which builds the model and returns the adjusted targets.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param sample_weight:
            The (optional) array of sample weights.

        :return:
            The vector of adjusted targets, potentially with a dictionary of additional information.
        """
        assert self._macs is not None, "No reference to the MACS object encapsulating the Master"
        self.backend.build()
        v = self.build(x=x, y=y)
        y_loss = self.y_loss(self.backend, numeric_variables=y, model_variables=v, sample_weight=sample_weight)
        # if macs is not fitted (i.e., we are in the first 'projection' iteration), there is no need to choose a
        # strategy since both of them will collapse to the same formulation, namely minimizing the y_loss only,
        # otherwise we use the beta step if either alpha is None, or beta is not None and use_beta() returns true
        if self._macs.fitted:
            p = self._macs.predict(x).reshape(v.shape)
            p_loss = self.p_loss(self.backend, numeric_variables=p, model_variables=v, sample_weight=sample_weight)
            # leave use_beta at the end to avoid function evaluation if one of the two conditions hold before
            if self.alpha is None or (self.beta is not None and self.use_beta(x=x, y=y, p=p)):
                beta = self.beta(macs=self._macs, x=x, y=y, p=p)
                self._log_stats(use_beta=1, beta=beta)
                self.backend.add_constraint(constraint=p_loss <= beta)
                self.backend.minimize(cost=y_loss)
            else:
                alpha = self.alpha(macs=self._macs, x=x, y=y, p=p)
                self._log_stats(use_beta=0, alpha=alpha)
                self.backend.minimize(cost=y_loss + (1.0 / alpha) * p_loss)
        else:
            p_loss = None
            self.backend.minimize(cost=y_loss)
        # if the problem is infeasible return no None, otherwise log stats and return the adjusted labels
        adjusted = None
        if self.backend.solve().solution is not None:
            adjusted = self.backend.get_values(expressions=v)
            self._log_stats(
                y_loss=self.backend.get_value(y_loss),
                p_loss=None if p_loss is None else self.backend.get_value(p_loss),
                objective=self.backend.get_objective()
            )
        self.backend.clear()
        return adjusted



"""Basic Loss Interface."""
from typing import Optional, Any, Callable, Union

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util.errors import not_implemented_message


class Loss:
    """Basic interface for a Moving Targets Master Loss."""

    def __init__(self, name: str):
        """
        :param name:
            The name of the loss.
        """

        self.__name__: str = name
        """The name of the loss."""

    def __call__(self,
                 backend: Backend,
                 numeric_variables: np.ndarray,
                 model_variables: np.ndarray,
                 sample_weight: Optional[np.ndarray] = None) -> Any:
        """Computes the loss value.

        :param backend:
            The `Backend` instance used to compute the loss.

        :param numeric_variables:
            The array of numeric ground truths.

        :param model_variables:
            The array of backend variables.

        :param sample_weight:
            The (optional) array of sample weights.

        :return:
            Either a single value/expression or an array of values/expressions.
        """
        raise NotImplementedError(not_implemented_message(name='__call__'))


class WeightedLoss(Loss):
    """Abstract class for a Moving Targets Master Loss which already handles Sample Weights."""

    def __init__(self, aggregation: Union[str, Callable], name: str):
        """
        :param aggregation:
            The kind of aggregation needed or a `Callable` function of type f(<backend>, <losses>) -> <result>.

            If the aggregation is 'sum', computes the sum of all losses over all dimensions.
            If the aggregation is 'mean', computes the mean of all losses over all dimensions.
            If the aggregation is 'mean_of_sums', aggregates each column (feature) via sum, then computes the mean.
            If the aggregation is 'sum_of_means', aggregates each column (feature) via mean, then computes the sum.

        :param name:
            The name of the loss.
        """
        if aggregation == 'sum':
            aggregation = lambda backend, losses: backend.sum(losses)
        elif aggregation == 'mean':
            aggregation = lambda backend, losses: backend.sum(losses) / losses.size
        elif aggregation == 'mean_of_sums':
            # if losses is a one-dimensional vector, this corresponds to losses.sum()
            # if losses is a two-dimensional vector, this corresponds to losses.mean(axis=0).sum()
            # in both cases, this can be achieved by computing losses.sum() / losses.shape[1]
            aggregation = lambda backend, losses: backend.sum(losses) / (1 if losses.ndim == 1 else losses.shape[1])
        elif aggregation == 'sum_of_means':
            # if losses is a one-dimensional vector, this corresponds to losses.mean()
            # if losses is a two-dimensional vector, this corresponds to losses.sum(axis=1).mean()
            # in both cases, this can be achieved by computing losses.sum() / losses.shape[0]
            aggregation = lambda backend, losses: backend.sum(losses) / losses.shape[0]
        elif not isinstance(aggregation, Callable):
            raise AssertionError(f"Unsupported aggregation strategy '{aggregation}'")

        super(WeightedLoss, self).__init__(name=name)

        self.aggregation: Callable = aggregation
        """The aggregation strategy."""

    def __call__(self,
                 backend: Backend,
                 numeric_variables: np.ndarray,
                 model_variables: np.ndarray,
                 sample_weight: Optional[np.ndarray] = None) -> Any:
        """Core method used to compute the master loss.

        :param backend:
            The `Backend` instance.

        :param numeric_variables:
            The real-number variables (i.e., the ground truths).

        :param model_variables:
            The model variables (i.e., the predictions).

        :param sample_weight:
            The sample weights associated to each sample.

        :return:
            The loss expression.
        """
        losses = self._losses(backend=backend, numeric_variables=numeric_variables, model_variables=model_variables)
        if sample_weight is not None:
            # normalize weights
            sample_weight = len(sample_weight) * np.array(sample_weight) / np.sum(sample_weight)
            # handle bi-dimensional outputs
            if losses.ndim == 2:
                sample_weight = np.repeat(sample_weight.reshape((-1, 1)), repeats=losses.shape[1], axis=1)
            # multiply partial loss per respective weight
            losses = sample_weight * losses
        return self.aggregation(backend, losses)

    def _losses(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        """Computes the partial losses computed over the pairs of variables.

        :param backend:
            The `Backend` instance.

        :param numeric_variables:
            The real-number variables (i.e., the ground truths).

        :param model_variables:
            The model variables (i.e., the predictions).

        :return:
            The array of partial losses computed over the pairs of variables.
        """
        raise NotImplementedError(not_implemented_message(name='_losses'))



from typing import Union

import numpy as np

from moving_targets.masters.backends import NumpyBackend
from moving_targets.masters.optimizers.optimizer import Optimizer
from moving_targets.util import probabilities
from moving_targets.util.errors import not_implemented_message


class BetaSatisfiability(Optimizer):
    """Dynamic Optimizer which should be used to control the behaviour of beta when dealing with satisfiability.

    Indeed, when using the beta step, it may be necessary to adopt an adaptive strategy in order to avoid the MACS
    process to stop due to the constraint 'p_loss <= beta' leading to infeasibility, since there are some cases in
    which the constraint satisfiability does not imply a null p_loss due to, e.g., (1) the usage of class probabilities
    instead of class targets in classification problems, in fact, since most of the classification losses (apart from
    HammingDistance) use class probabilities while the satisfiability is usually computed on class targets instead,
    there will always be a minimal amount of error due to the loss discrepancy between binary and continuous variables,
    or (2) the presence of some lower/upper bounds in regression problems that are not considered in the constraint
    satisfiability definition but are required by the model variables definition, which may happen for example if we
    know that our targets must be non-negative (thus we force our lower bound to be 0) but the learner might return
    some negative predictions that will introduce an error in the p_loss.

    Still, this minimal error can be measured as the loss between the predictions and their "model" representation,
    which involves variable bounds in the case of regression and the handling of targets and probabilities in the case
    of classification. This latter part is delegated to this method, which must return the accordingly constrained
    variables in order to let the master compute this error relying on its inner loss.
    """

    def __init__(self, base: Union[float, Optimizer]):
        """
        :param base:
            Either a fixed floating point value representing the initial value for the hyper-parameter to optimize, or
            a wrapped custom optimizer which returns a dynamic value to reduce by the given factor after each iteration.

        """
        super(BetaSatisfiability, self).__init__(base=base)

    def _expected_variables(self, macs, x, y: np.ndarray, p: np.ndarray) -> np.ndarray:
        """Computes the expected "model" representation of the predictions.

        :param macs:
            The `MACS` instance encapsulating the master.

        :param x:
            The training samples.

        :param y:
            The training labels.

        :param p:
            The `Learner` predictions.

        :return:
            An array representing the expected "model" representation of the predictions.
        """
        raise NotImplementedError(not_implemented_message(name='_expected_variables'))

    def __call__(self, macs, x, y: np.ndarray, p: np.ndarray) -> float:
        value = super(BetaSatisfiability, self).__call__(macs, x, y, p)
        if p is not None:
            # in order to implement the adaptive strategy used to avoid model infeasibility due to the beta constraint
            # we rely on the NumpyBackend instance which can compute the value of the given loss between two numpy
            # arrays; this value will be our minimal error to be added to beta
            v = self._expected_variables(macs, x, y, p)
            value += macs.master.p_loss(backend=NumpyBackend(), numeric_variables=p, model_variables=v)
        return value


class BetaClassSatisfiability(BetaSatisfiability):
    """Beta Satisfiability Optimizer which adjusts the hyper-parameter value by computing the minimal error amount
    between the predicted probabilities and their class representation.

    This may be useful when dealing with a classification loss that uses predicted probabilities while checking for the
    model satisfiability on the predicted classes, since in that case there will be an amount of error in the p_loss
    that is not dependent from the constraint satisfaction.
    """

    def __init__(self, base: Union[float, Optimizer], labelling: bool = False):
        """
        :param base:
            Either a fixed floating point value representing the initial value for the hyper-parameter to optimize, or
            a wrapped custom optimizer which returns a dynamic value to reduce by the given factor after each iteration.

        :param labelling:
            Whether the master represents a labelling or a classification task.
        """
        super(BetaClassSatisfiability, self).__init__(base=base)

        self.labelling: bool = labelling
        """Whether the master represents a labelling or a classification task."""

    def _expected_variables(self, macs, x, y: np.ndarray, p: np.ndarray) -> np.ndarray:
        # the classes are obtained using the 'get_classes' method
        c = probabilities.get_classes(prob=p, labelling=self.labelling)
        # if we are dealing with a classification task, the classes will be one-dimensional, thus we onehot encode them
        if c.ndim == 1:
            onehot_binary = p.ndim == 2
            return probabilities.get_onehot(c, classes=p.shape[1] if onehot_binary else 2, onehot_binary=onehot_binary)
        else:
            return c


class BetaBoundedSatisfiability(BetaSatisfiability):
    """Beta Satisfiability Optimizer which adjusts the hyper-parameter value by computing the minimal error amount
    between the model predictions and their constrained representation which must satisfy the lower and upper bounds.

    This may be useful when dealing with regression losses that is unconstrained (due to the employed machine learning
    model) while putting some lower/upper bounds constraints in the solver, since in that case there will be an amount
    of error in the p_loss that is not dependent from the constraint satisfaction.
    """

    def __init__(self, base: Union[float, Optimizer], lb: float, ub: float):
        """
        :param base:
            Either a fixed floating point value representing the initial value for the hyper-parameter to optimize, or
            a wrapped custom optimizer which returns a dynamic value to reduce by the given factor after each iteration.

        :param lb:
            The model variables lower bounds.

        :param ub:
            The model variables upper bounds.
        """
        super(BetaBoundedSatisfiability, self).__init__(base=base)

        self.lb: float = lb
        """The model variables lower bounds."""

        self.ub: float = ub
        """The model variables upper bounds."""

    def _expected_variables(self, macs, x, y: np.ndarray, p: np.ndarray) -> np.ndarray:
        # the constrained representation is obtained by projecting the predictions in the feasible space via clipping
        return p.clip(min=self.lb, max=self.ub)



"""Weights&Biases Callback"""

from typing import Dict, Optional

import numpy as np

from moving_targets.callbacks.logger import Logger
from moving_targets.util.errors import MissingDependencyError
from moving_targets.util.typing import Dataset


class WandBLogger(Logger):
    """Logs the training information on a Weights&Biases instance."""

    def __init__(self, project: str, entity: str, run_name: str, **wandb_kwargs):
        """
        :param project:
            Weights&Biases project name.

        :param entity:
            Weights&Biases entity name.

        :param run_name:
            Weights&Biases run name.

        :param wandb_kwargs:
            Weights&Biases run configuration.
        """
        super(WandBLogger, self).__init__()

        try:
            import wandb
            self._wandb = wandb
            """The lazily imported wandb instance."""
        except ModuleNotFoundError:
            raise MissingDependencyError(package='wandb')

        self._project: str = project
        """The Weights&Biases project name."""

        self._entity: str = entity
        """The Weights&Biases entity name."""

        self._run_name: str = run_name
        """The Weights&Biases run name."""

        self.config: Dict = wandb_kwargs
        """The Weights&Biases run configuration."""

    def on_process_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._wandb.init(project=self._project, entity=self._entity, name=self._run_name, config=self.config)

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._wandb.log({k: self._cache[k] for k in sorted(self._cache)})
        self._cache = {}

    def on_process_end(self, macs, val_data: Optional[Dataset]):
        self._wandb.finish()



"""History Callback"""
import warnings
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from moving_targets.callbacks.logger import Logger
from moving_targets.util.typing import Dataset


class History(Logger):
    """Collects the training information and eventually plots it."""

    def __init__(self):
        """"""
        super(History, self).__init__()

        self._history: List = []
        """An internal list which will be used to create the final DataFrame object at the end of the training."""

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # Creates a dataframe from the cached items and appends them to a list indexed by iteration.
        self._history.append(pd.DataFrame([self._cache.values()], columns=self._cache.keys()))
        self._cache = {}

    def on_process_end(self, macs, val_data: Optional[Dataset]):
        # Creates a single dataframe by concatenating the sub-dataframes from each iteration.
        if len(self._history) > 0:
            self._history = pd.concat(self._history, ignore_index=True)

    def plot(self,
             features: Optional[List[str]] = None,
             num_subplots: Union[str, int] = '/',
             orient_rows: bool = False,
             tight_layout=True,
             figsize=(16, 9),
             **plt_kwargs):
        """Plots the training information which were previously collected in a dataframe.

        :param features:
            List of strings representing the names of the columns to plot. If None, plots all the columns.

        :param num_subplots:
            float of row/columns in the final subplot, respectively to the value of the orient_rows argument.
            If a string is passed, that is used to display the subplots according to a common prefix (e.g.,
            if num_subplots = '/' and the features are ['trn/loss', 'trn/metric', 'val/loss', 'val/metric'],
            then there will be a row/column for the 'trn' prefix and another one for the 'val' prefix).

        :param orient_rows:
            Whether to orient the plots by column or by row. This influences also the 'num_subplots' parameter since
            when 'orient_rows' is set to True the 'num_subplots' indicates the number of rows while when 'orient_rows'
            is set to False the 'num_subplots' indicates the number of columns.

        :param tight_layout:
            Matplotlib `figure()` argument.

        :param figsize:
            Matplotlib `figure()` argument.

        :param plt_kwargs:
            Additional plot arguments to be passed to `figure()`.
        """
        if not isinstance(self._history, pd.DataFrame):
            warnings.warn('Process did not end correctly, therefore no dataframe can be plotted')
            return
        # HANDLE COLUMNS TO BE PLOTTED (FEATURES)
        #   1. we use the given columns if explicitly passed, otherwise we use the df columns
        #   2. the features are then obtained as those who are a subset of the number data type
        features = self._history.columns if features is None else features
        features = [f for f in features if f is None or np.issubdtype(self._history[f].dtype, np.number)]
        # HANDLE THE SUBPLOT POSITIONS
        #   1. if the given number of subplots is an integer, it arranges the features by taking them in groups of
        #      <num_subplots>, otherwise it means that the argument is a prefix separator thus it arranges the features
        #      by constructing a dictionary of features sharing the prefix and taking them as a group
        #   2. computes the dimensions of the newly arranged features list as the number of rows and the length of the
        #      longest column, then creates a numpy array of None values having that shape and eventually fills it
        #   3. if 'orient_column' is False, transpose the array of positions
        if isinstance(num_subplots, int):
            features = [features[i::num_subplots] for i in range(num_subplots)]
            # the behaviour here is inverted since we are taking the features by <num_subplots> steps
            orient_rows = not orient_rows
        else:
            prefixes = {}
            for column in features:
                prefix = column.split(num_subplots)[0]
                if prefix in prefixes:
                    prefixes[prefix].append(column)
                else:
                    prefixes[prefix] = [column]
            features = [v for v in prefixes.values()]
        num_rows = len(features)
        num_cols = np.max([len(row) for row in features])
        positions = np.array([[None] * num_cols] * num_rows)
        for i, row in enumerate(features):
            # filling up to len(row) is necessary since some rows may have less features than the maximum one
            positions[i, :len(row)] = row
        positions = positions if orient_rows else positions.transpose()
        # PLOT EACH FEATURE IN A SUBPLOT
        #   1. iterates by rows and columns
        #   2. if the feature name is not None, it plots the data
        ax = None
        num_rows, num_cols = positions.shape
        plt.figure(figsize=figsize, tight_layout=tight_layout, **plt_kwargs)
        for idx, feature in enumerate(positions.flatten()):
            if feature is not None:
                ax = plt.subplot(num_rows, num_cols, idx + 1, sharex=ax)
                x, y = self._history.index, self._history[feature]
                plt.plot(x, y)
                plt.scatter(x, y)
                ax.set(title=feature, xlabel='', ylabel='')
                # fix integer x_ticks based on autogenerated ones in order to avoid superimposed values
                ticks = np.unique(ax.get_xticks().round().astype(int))
                ax.set_xticks([t for t in ticks if t in range(x.min(), x.max() + 1)])
        plt.show()



"""Basic Regression Master."""

from abc import ABC
from typing import Union, List

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.masters.backends import get_backend
from moving_targets.masters.losses import RegressionLoss, Loss, aliases
from moving_targets.masters.losses import SAE, SSE, MAE, MSE
from moving_targets.masters.master import Master
from moving_targets.masters.optimizers import BetaBoundedSatisfiability
from moving_targets.masters.optimizers import Optimizer


class RegressionMaster(Master, ABC):
    """An abstract Moving Targets Master for Regression Tasks."""

    def __init__(self,
                 backend: Union[str, Backend],
                 lb: float = -float('inf'),
                 ub: float = float('inf'),
                 alpha: Union[None, float, Optimizer] = 1.0,
                 beta: Union[None, float, Optimizer] = None,
                 y_loss: Union[str, RegressionLoss] = 'mse',
                 p_loss: Union[str, RegressionLoss] = 'mse',
                 stats: Union[bool, List[str]] = False):
        """
        :param backend:
            The `Backend` instance encapsulating the optimization solver.

        :param lb:
            The model variables lower bounds.

        :param ub:
            The model variables upper bounds.

        :param alpha:
            Either a constant alpha value, an alpha optimizer, or None if no alpha step is wanted.

        :param beta:
            Either a constant beta value, a beta optimizer, or None if no beta step is wanted.

        :param y_loss:
            A `Loss` instance used to compute the y_loss.

        :param p_loss:
            A `Loss` instance used to compute the p_loss.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in ['alpha',
            'beta', 'use_beta', 'y_loss', 'p_loss', 'objective'] whose statistics must be logged.
        """

        self.lb = lb
        """The model variables lower bounds."""

        self.ub = ub
        """The model variables upper bounds."""

        super(RegressionMaster, self).__init__(
            backend=get_backend(backend) if isinstance(backend, str) else backend,
            y_loss=self._get_loss(loss=y_loss),
            p_loss=self._get_loss(loss=p_loss),
            alpha=None if alpha is None else Optimizer(base=alpha),
            beta=None if beta is None else BetaBoundedSatisfiability(base=beta, lb=lb, ub=ub),
            stats=stats
        )

    def build(self, x, y: np.ndarray) -> np.ndarray:
        return self.backend.add_continuous_variables(*y.shape, lb=self.lb, ub=self.ub, name='y')

    @staticmethod
    def _get_loss(loss: Union[str, Loss]) -> Loss:
        """Handles the given loss by checking that it is both known and supported.

        :param loss:
            The given loss.

        :return:
            A `Loss` instance.

        :raise `AssertionError`:
            If either the loss alias is unknown or does not represent a regression loss.
        """
        if isinstance(loss, Loss):
            # if the loss is already a loss instance we simply return it
            return loss
        else:
            # otherwise we retrieve the class from the alias then check if it is both known and supported
            class_type = aliases.get(loss)
            assert class_type is not None, f"Unknown loss alias '{loss}'"
            assert class_type in [SAE, SSE, MAE, MSE], f"Loss type {class_type.__name__} is not a regression loss."
            return class_type(name=loss)



"""Basic Classification Master."""
from abc import ABC
from typing import Set, Callable, Optional
from typing import Union, List

import numpy as np

from moving_targets.masters.backends import Backend, get_backend
from moving_targets.masters.losses import SAE, SSE, MAE, MSE, Loss, aliases, HammingDistance, CrossEntropy, \
    ReversedCrossEntropy, SymmetricCrossEntropy
from moving_targets.masters.master import Master
from moving_targets.masters.optimizers import Optimizer, BetaClassSatisfiability
from moving_targets.util import probabilities
from moving_targets.util.typing import Dataset


class ClassificationMaster(Master, ABC):
    """An abstract Moving Targets Master for Classification Tasks."""

    TASKS: Set[str] = {'auto', 'binary', 'multiclass', 'multilabel'}
    """Set of accepted tasks."""

    VTYPES: Set[str] = {'auto', 'discrete', 'continuous'}
    """Set of accepted vtypes."""

    RTYPES: Set[str] = {'class', 'probability'}
    """Set of accepted rtypes."""

    def __init__(self,
                 backend: Union[str, Backend],
                 alpha: Union[None, float, Optimizer] = 1.0,
                 beta: Union[None, float, Optimizer] = None,
                 y_loss: Union[str, Loss] = 'hd',
                 p_loss: Union[str, Loss] = 'mse',
                 task: str = 'auto',
                 vtype: str = 'auto',
                 rtype: str = 'class',
                 stats: Union[bool, List[str]] = False):
        """
        :param backend:
            The `Backend` instance encapsulating the optimization solver.

        :param alpha:
            Either a constant alpha value, an alpha optimizer, or None if no alpha step is wanted.

        :param beta:
            Either a constant beta value, a beta optimizer, or None if no beta step is wanted.

        :param y_loss:
            A `Loss` instance used to compute the y_loss.

        :param p_loss:
            A `Loss` instance used to compute the p_loss.

        :param task:
            The kind of classification task, which can be either 'binary', 'multiclass', 'multilabel', or 'auto'.

            If 'binary' is chosen, the model variables will be stored in a one-dimensional vector.
            If 'multiclass' is chosen, the model variables will be stored in a matrix so that each row sum up to one.
            If 'multilabel' is chosen, the model variables will be stored in a matrix without any summing constraint.
            If 'auto' is chosen, it will choose between 'binary' or 'multiclass' depending on the number of classes
            (if you need to handle a multilabel classification task, you should explicitly declare it).

        :param vtype:
            The model variables vtype, either 'continuous' or 'discrete', or 'auto' for automatic type inference
            depending on the given losses.

        :param rtype:
            The return type of the adjusted targets, either 'class' to get class targets, or 'probability' to get class
            probabilities. Please notice that when returning probabilities, the learner may be adjusted accordingly,
            since some learners do not accept continuous targets and/or bi-dimensional targets.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in ['alpha',
            'beta', 'use_beta', 'y_loss', 'p_loss', 'objective'] whose statistics must be logged.

        :raise `TypeError`:
            If two losses cannot handle the same (forced) variable types or if a loss cannot handle an externally
            forced variable type.
        """
        assert task in self.TASKS, f"'task' should be in {self.TASKS}, got '{task}'"
        assert vtype in self.VTYPES, f"'vtype' should be in {self.VTYPES}, got '{vtype}'"
        assert rtype in self.RTYPES, f"'rtype' should be in {self.RTYPES}, got '{rtype}'"
        # handle losses aliases and automatic type inference, e.g.:
        #   > if we have (y_loss = 'hd', p_loss = 'mse', vtype = 'auto') we will set binary = True since hamming
        #     distance can only handle binary targets
        #   > similarly, if we have (y_loss = 'hd', p_loss = 'rce', vtype = 'auto'), because even though reversed
        #     crossentropy prefers continuous targets, hamming distance forces them to be binary
        #   > on the contrary if, for example, we have (y_loss = 'mse', p_loss = 'rce', vtype = 'auto'), then we will
        #     set binary = False since reversed crossentropy prefers continuous targets and mean squared error does not
        #     force any variable type
        #   > however, if we get have (y_loss = 'mse', p_loss = 'rce', vtype = 'binary') then binary = True because
        #     the variables type is externally forced
        #   > moreover, if a given loss is already a Loss instance, this is supposed to expect binary targets if the
        #     variable types are not forced by the user
        #   > finally, if we have something like (y_loss = 'hd', p_loss = 'rce', vtype = 'continuous'), then we will
        #     raise a TypeError because targets are forced to be continuous but hamming distance cannot handle them
        y_loss, p_loss = self._get_loss(loss=y_loss, task=task), self._get_loss(loss=p_loss, task=task)
        force_binary = vtype == 'discrete' or y_loss.force_binary or p_loss.force_binary
        force_continuous = vtype == 'continuous' or y_loss.force_continuous or p_loss.force_continuous
        if not force_binary or not force_continuous:
            # if either one between binary and continuous is not forced, we can proceed and set the correct type
            binary = force_binary or (y_loss.prefer_binary and p_loss.prefer_binary)
            y_loss, p_loss = y_loss.loss_fn(b=binary), p_loss.loss_fn(b=binary)
            vtype = 'binary' if binary else 'continuous'
        elif vtype == 'auto':
            # if vtype is 'auto', it means that the two losses forced two different types, thus they cannot go together
            raise TypeError(f"y_loss '{y_loss}' and p_loss '{p_loss}' cannot handle the same vtypes")
        else:
            # otherwise, it means that the vtype that was externally forced cannot be handled by some of the losses
            raise TypeError(f"Either y_loss '{y_loss}' or p_loss '{p_loss}' cannot deal with vtype='{vtype}'")

        self.task: str = task
        """The kind of classification task."""

        self.vtype: str = vtype
        """The model variables vtype."""

        self.rtype: str = rtype
        """The model variables vtype."""

        super(ClassificationMaster, self).__init__(
            backend=get_backend(backend) if isinstance(backend, str) else backend,
            y_loss=y_loss,
            p_loss=p_loss,
            alpha=None if alpha is None else Optimizer(base=alpha),
            beta=None if beta is None else BetaClassSatisfiability(base=beta, labelling=task == 'labelling'),
            stats=stats
        )

    def on_process_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # set task in case of automatic task detection based on the number of classes
        if self.task == 'auto':
            assert y.squeeze().ndim == 1, "'auto' task can handle 1d targets only, use 'multilabel' for 2d targets"
            self.task = 'binary' if probabilities.count_classes(vector=y) == 2 else 'multiclass'

    def build(self, x, y: np.ndarray) -> np.ndarray:
        # we can simply build variables by shape (the y targets will be already onehot encoded for multiclass tasks)
        variables = self.backend.add_variables(*y.shape, vtype=self.vtype, lb=0, ub=1, name='y')
        if self.task == 'multiclass':
            # if we are dealing with multiclass classification, we create a bi-dimensional matrix and constraint the
            # variables in each row to sum up to one since we want class probabilities
            self.backend.add_constraints([self.backend.sum(row) == 1 for row in variables])
        return variables

    def adjust_targets(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
        if self.task == 'multiclass':
            # for multiclass classification tasks, we need to acknowledge that MT losses must work with both original
            # targets and predictions thus, since predictions usually come as 2d floating point array of probabilities
            # while class targets usually come as an integer vector, we must onehot encode the latter for compatibility
            # (this will not be necessary for multilabel classification since the original targets will be already 2d)
            y = probabilities.get_onehot(y, onehot_binary=True)
            labelling = False
        else:
            # for binary tasks, the parameter 'task' of get_discrete() is ignored, thus we can simply pass 'labelling'
            labelling = True
        adjusted = super(ClassificationMaster, self).adjust_targets(x, y, sample_weight)
        return probabilities.get_classes(adjusted, labelling=labelling) if self.rtype == 'class' else adjusted

    class _LossInfo:
        """Data class containing information about a loss."""

        def __init__(self,
                     loss_fn: Callable,
                     force_binary: bool = False,
                     force_continuous: bool = False,
                     prefer_binary: bool = True):
            """
            :param loss_fn:
                A callable function of type f(<binary>) -> <Loss> that builds the loss instance.

            :param force_binary:
                Whether the loss can handle only binary variables or not.

            :param force_continuous:
                Whether the loss can handle only continuous variables or not.

            :param prefer_binary:
                Whether the loss prefers to handle binary variables by default or not.
            """
            self.loss_fn: Callable = loss_fn
            """A callable function of type f(<binary>) -> <Loss> that builds the loss instance."""

            self.force_binary: bool = force_binary
            """Whether the loss can handle only binary variables or not."""

            self.force_continuous: bool = force_continuous
            """Whether the loss can handle only continuous variables or not."""

            self.prefer_binary: bool = prefer_binary
            """Whether the loss prefers to handle binary variables by default or not."""

    @classmethod
    def _get_loss(cls, loss: Union[str, Loss], task: str) -> _LossInfo:
        """Handles the given loss by checking, then returns its respective `_LossInfo` instance.

        :param loss:
            The given loss.

        :param task:
            The given task.

        :return:
            A `Loss` instance.

        :raise `AssertionError`:
            If the loss alias is unknown.
        """
        if isinstance(loss, str):
            # if we get passed a string, we retrieve the class type and build a correct instance accordingly
            class_type = aliases.get(loss)
            if class_type == HammingDistance:
                # hamming distance can handle only binary targets, also, for this loss we need to specify the task
                return cls._LossInfo(loss_fn=lambda b: HammingDistance(labelling=task == 'multilabel', name=loss),
                                     force_binary=True)
            elif class_type == CrossEntropy:
                # crossentropy can handle only binary targets
                return cls._LossInfo(loss_fn=lambda b: CrossEntropy(name=loss), force_binary=True)
            elif class_type in [ReversedCrossEntropy, SymmetricCrossEntropy]:
                # reversed and symmetric crossentropy prefer continuous targets but can handle both
                return cls._LossInfo(loss_fn=lambda b: class_type(binary=b, name=loss), prefer_binary=False)
            elif class_type in [SAE, SSE, MAE, MSE]:
                return cls._LossInfo(loss_fn=lambda b: class_type(binary=b, name=loss))
            else:
                raise AssertionError(f"Unknown loss alias '{loss}'")
        else:
            # otherwise, if the loss is already a loss instance, we ignore the <binary> parameter and return it
            return cls._LossInfo(loss_fn=lambda b: loss)



from typing import Callable

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.masters.losses.loss import WeightedLoss


class RegressionLoss(WeightedLoss):
    """Basic interface for a Moving Targets Regression Loss."""

    def __init__(self, norm: int, sum_samples: bool, sum_features: bool, binary: bool, name: str = 'absolute_errors'):
        """
        :param norm:
            Either '1' for absolute errors or '2' for squared errors.

        :param sum_samples:
            Whether to aggregate the partial losses via sum or via mean.

        :param sum_features:
            Whether to sum the results of each feature or to get their mean value, in case of multiple features.

        :param binary:
            Whether the model variables are expected to be binary or not.

            Knowing that the model variables are binary allows to deal with error computation in a faster way since it
            will lead to a simplified scenario in which the model variables can only have an integer value in {0, 1},
            while the numeric variables can only have a continuous value in [0, 1].

        :param name:
            The name of the loss.
        """
        assert norm in [1, 2], f"'norm' should be either '1' for absolute errors or '2' for squared errors, got {norm}"
        # handle aggregation by considering strategy to aggregate both samples and features
        if sum_samples:
            aggregation = 'sum' if sum_features else 'mean_of_sums'
        else:
            aggregation = 'sum_of_means' if sum_features else 'mean'
        super(RegressionLoss, self).__init__(aggregation=aggregation, name=name)

        if binary:
            # given the model variables <m> and the numeric variables <n>, in this scenario we have that:
            #
            # > abs(m, n) = m * (1 - n) + (1 - m) * n =
            #             = m - 2mn + n
            #
            # > sqr(m, n) = [m * (1 - n) + (1 - m) * n]^2 =
            #             = [m * (1 - n)]^2        + [(1 - m) * n]^2	    - 2 * [m * (1 - n) * (1 - m) * n] =
            #             = [m^2 * (1 - 2n + n^2)] + [n^2 * (1 - 2m + m^2)] - 2 * [m * (1 - m) * (1 - n) * n] =
            #             = [m - 2mn + mn^2] 	   + [n^2 - 2mn^2 + mn^2]   - 2 * [(m - m) * (n - n^2)] = --> as m^2 = m
            #             = [m - 2mn + mn^2]       + [n^2 - mn^2]           - 2 * [0] =
            #             = m - 2mn + n^2
            #
            # thus, since the absolute values represent norm 1 and the squared values represent norm 2, we can rewrite
            # the formulae as "m - 2mn + n^{norm}", which is a linear expressions that can be handled in a faster way
            strategy = lambda b, n, m: m - 2 * m * n + n ** norm
        elif norm == 1:
            # if we have continuous targets with norm one, we apply the absolute value of the differences
            strategy = lambda b, n, m: b.abs(m - n)
        else:
            # if we have continuous targets with norm one, we square the differences
            strategy = lambda b, n, m: b.square(m - n)

        self._strategy: Callable = strategy
        """The norm strategy."""

    def _losses(self, backend: Backend, numeric_variables: np.ndarray, model_variables: np.ndarray) -> np.ndarray:
        return self._strategy(b=backend, n=numeric_variables, m=model_variables)


class SAE(RegressionLoss):
    """Sum of Absolute Errors Loss."""

    def __init__(self, binary: bool = False, sum_features: bool = True, name: str = 'sum_of_absolute_errors'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            Knowing that the model variables are binary allows to deal with error computation in a faster way since it
            will lead to a simplified scenario in which the model variables can only have an integer value in {0, 1},
            while the numeric variables can only have a continuous value in [0, 1].

        :param sum_features:
            Whether to sum the results of each feature or to get their mean value, in case of multiple features.

        :param name:
            The name of the loss.
        """
        super(SAE, self).__init__(norm=1, sum_samples=True, sum_features=sum_features, binary=binary, name=name)


class SSE(RegressionLoss):
    """Sum of Squared Errors Loss."""

    def __init__(self, binary: bool = False, sum_features: bool = True, name: str = 'sum_of_squared_errors'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            Knowing that the model variables are binary allows to deal with error computation in a faster way since it
            will lead to a simplified scenario in which the model variables can only have an integer value in {0, 1},
            while the numeric variables can only have a continuous value in [0, 1].

        :param sum_features:
            Whether to sum the results of each feature or to get their mean value, in case of multiple features.

        :param name:
            The name of the loss.
        """
        super(SSE, self).__init__(norm=2, sum_samples=True, sum_features=sum_features, binary=binary, name=name)


class MAE(RegressionLoss):
    """Mean Absolute Error Loss."""

    def __init__(self, binary: bool = False, sum_features: bool = True, name: str = 'mean_absolute_error'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            Knowing that the model variables are binary allows to deal with error computation in a faster way since it
            will lead to a simplified scenario in which the model variables can only have an integer value in {0, 1},
            while the numeric variables can only have a continuous value in [0, 1].

        :param sum_features:
            Whether to sum the results of each feature or to get their mean value, in case of multiple features.

        :param name:
            The name of the loss.
        """
        super(MAE, self).__init__(norm=1, sum_samples=False, sum_features=sum_features, binary=binary, name=name)


class MSE(RegressionLoss):
    """Mean Squared Error Loss."""

    def __init__(self, binary: bool = False, sum_features: bool = True, name: str = 'mean_squared_error'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            Knowing that the model variables are binary allows to deal with error computation in a faster way since it
            will lead to a simplified scenario in which the model variables can only have an integer value in {0, 1},
            while the numeric variables can only have a continuous value in [0, 1].

        :param sum_features:
            Whether to sum the results of each feature or to get their mean value, in case of multiple features.

        :param name:
            The name of the loss.
        """
        super(MSE, self).__init__(norm=2, sum_samples=False, sum_features=sum_features, binary=binary, name=name)



"""Custom scalers."""

from typing import Union, Dict, List, Optional, Any

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelBinarizer


class Scaler:
    """Custom scaler that is able to treat each feature separately."""

    def __init__(self, *methods: str, default_method: str = 'none', **custom_methods: str):
        """
        :param methods:
            If a single value is passed, all the features are scaled with the same method, otherwise, if a list or a
            dictionary is passed, each feature is scaled with the respective method.

        :param default_method:
            The default scaling which is used for features for which an explicit method is not passed.

            Admitted methods are (1) 'std' to standardize the data, (2) 'norm' to normalize the data, (3) 'onehot' to
            onehot encode categorical features, and (4) 'none' to perform no scaling.
        """
        super(Scaler, self).__init__()
        assert default_method == 'none' or len(methods) == 0, "if the default method is not None, custom scaling " \
                                                              "methods for specific features must be passed via " \
                                                              "keyword (i.e., the '*methods' parameter must be empty)"
        assert len(custom_methods) == 0 or len(methods) <= 1, "if at least one custom scaling method for specific " \
                                                              "features is passed, no unnamed method must be passed " \
                                                              "as well (i.e., the '*methods' parameter must be empty)"

        # if a single method is passed (and default_method is known to be None), then we consider it the default one
        self.default_method: str = methods[0] if len(methods) == 1 else default_method
        """The default scaling method."""

        # if more than a single method is passed we use them, otherwise we use the custom_methods dictionary
        self.methods: Union[List[str], Dict[str, str]] = list(methods) if len(methods) > 1 else custom_methods
        """The custom scaling methods."""

        self._dtypes: Any = None
        """The input data types."""

        self.input_columns: List[Any] = []
        """The list of input column names."""

        self.output_columns: List[Any] = []
        """The list of output column names (this is useful for onehot encoding)."""

        self._onehot: Dict[Any, LabelBinarizer] = {}
        """The dictionary of onehot encoders indexed by column name/number."""

        self._translation: Optional[np.ndarray] = None
        """The translation vector."""

        self._scaling: Optional[np.ndarray] = None
        """The scaling vector."""

    def fit(self, data) -> Any:
        """Fits the scaler parameters.

        :param data:
            The matrix/dataframe of samples.

        :return:
            The scaler itself.
        """
        # handle input, then clear and fit metadata
        self.output_columns = []
        self.input_columns = []
        self._onehot = {}
        self._dtypes, data = (data.dtypes, data) if isinstance(data, pd.DataFrame) else (data.dtype, pd.DataFrame(data))

        # handle methods:
        # > if self.methods is a list, pair the methods with the columns
        # > otherwise, use the custom methods when passed, or the default method otherwise
        if isinstance(self.methods, list):
            assert len(self.methods) == len(data.columns), "previously passed methods do not match data features length"
            methods = {c: m for m, c in zip(self.methods, data.columns)}
        else:
            methods = {c: self.methods.get(c) or self.default_method for c in data.columns}

        # compute factors
        translation, scaling = [], []
        self.input_columns = list(data.columns)
        for column, method in methods.items():
            if method == 'onehot':
                encoder = LabelBinarizer()
                encoder.fit(list(data[column]))
                if encoder.y_type_ == 'binary':
                    self.output_columns.append(f'{column}_{encoder.classes_[1]}')
                    translation.append(0)
                    scaling.append(1)
                else:
                    self.output_columns += [f'{column}_{c}' for c in encoder.classes_]
                    translation += [0] * len(encoder.classes_)
                    scaling += [1] * len(encoder.classes_)
                self._onehot[column] = encoder
            else:
                self.output_columns.append(column)
                values = data[column].astype(np.float64).values
                if method in ['std', 'standardize']:
                    translation.append(values.mean())
                    scaling.append(values.std())
                elif method in ['norm', 'normalize', 'minmax']:
                    translation.append(values.min())
                    scaling.append(values.max() - values.min())
                elif method == 'none':
                    translation.append(0)
                    scaling.append(1)
                else:
                    raise ValueError(f'Method {method} is not supported')

        # handle case with null scaling factor to avoid division by zero
        self._scaling = np.where([s != 0 for s in scaling], scaling, 1.0).astype(np.float64)
        self._translation = np.array(translation).astype(np.float64)
        return self

    def transform(self, data) -> Any:
        """Transforms the data according to the scaler parameters.

        :param data:
            The matrix/dataframe of samples.

        :return:
            The scaled data.
        """
        is_1d, is_numpy = data.ndim == 1, True if isinstance(data, np.ndarray) else False
        data = pd.DataFrame(data, columns=self.input_columns)
        # handle one hot encoding
        columns, dataframes = self.input_columns.copy(), [data]
        for column, encoder in self._onehot.items():
            # define class names
            if encoder.y_type_ == 'binary':
                class_names = [f'{column}_{encoder.classes_[1]}']
            else:
                class_names = [f'{column}_{c}' for c in encoder.classes_]
            # retrieve column index and replace it with the list of class names
            index = columns.index(column)
            columns = columns[:index] + class_names + ([] if len(columns) == index + 1 else columns[index + 1:])
            # compute onehot classes, create dataframe, and append it to the list
            classes = encoder.transform(list(data[column]))
            classes = pd.DataFrame(classes, index=data.index, columns=class_names)
            dataframes.append(classes)
        # build the full dataframe by appending the classes at the end and reordering the columns, then scale it
        data = pd.concat(dataframes, axis=1)[columns]
        data = (data.astype(np.float64) - self._translation) / self._scaling
        data = data.squeeze() if is_1d else data
        return data.values if is_numpy else data

    def fit_transform(self, data) -> Any:
        """Fits the scaler parameters.

        :param data:
            The matrix/dataframe of samples.

        :return:
            The scaled data.
        """
        return self.fit(data).transform(data)

    def inverse_transform(self, data) -> Any:
        """Inverts the scaling according to the scaler parameters.

        :param data:
            The previously scaled matrix/dataframe of samples.

        :return:
            The original data.
        """
        is_1d, is_numpy = data.ndim == 1, True if isinstance(data, np.ndarray) else False
        data = pd.DataFrame(data, columns=self.output_columns)
        data = (data.astype(np.float64) * self._scaling) + self._translation
        # handle one hot encoding
        columns = self.output_columns.copy()
        for column, encoder in self._onehot.items():
            # define class names
            if encoder.y_type_ == 'binary':
                class_names = [f'{column}_{encoder.classes_[1]}']
            else:
                class_names = [f'{column}_{c}' for c in encoder.classes_]
            # remove onehot class names from columns and add column name instead
            first_index = columns.index(class_names[0])
            last_index = first_index + len(class_names)
            columns = columns[:first_index] + [column] + columns[last_index:]
            # add new column in the dataframe for the categorical feature
            data[column] = encoder.inverse_transform(data[class_names].values)
        # select only the correct columns and assign dtypes
        data = data[columns].astype(self._dtypes)
        data = data.squeeze() if is_1d else data
        return data.values if is_numpy else data



"""Moving Targets Losses for Masters."""
from moving_targets.masters.losses.classification import ClassificationLoss, HammingDistance, CrossEntropy, \
    ReversedCrossEntropy, SymmetricCrossEntropy
from moving_targets.masters.losses.loss import Loss, WeightedLoss
from moving_targets.masters.losses.regression import RegressionLoss, SAE, SSE, MAE, MSE

aliases: dict = {
    # Sum of Absolute Errors
    'sae': SAE,
    'sum_of_absolute_errors': SAE,
    'sum of absolute errors': SAE,
    # Sum of Squared Errors
    'sse': SSE,
    'sum_of_squared_errors': SSE,
    'sum of squared errors': SSE,
    # Mean Absolute Error
    'mae': MAE,
    'mean_absolute_error': MAE,
    'mean absolute error': MAE,
    # Mean Squared Error
    'mse': MSE,
    'mean_squared_error': MSE,
    'mean squared error': MSE,
    # Hamming Distance
    'hd': HammingDistance,
    'hamming_distance': HammingDistance,
    'hamming distance': HammingDistance,
    'bh': HammingDistance,
    '_hamming': HammingDistance,
    ' hamming': HammingDistance,
    'ch': HammingDistance,
    'categorical_hamming': HammingDistance,
    'categorical hamming': HammingDistance,
    # CrossEntropy
    'ce': CrossEntropy,
    'crossentropy': CrossEntropy,
    'bce': CrossEntropy,
    'binary_crossentropy': CrossEntropy,
    'binary crossentropy': CrossEntropy,
    'cce': CrossEntropy,
    'categorical_crossentropy': CrossEntropy,
    'categorical crossentropy': CrossEntropy,
    'll': CrossEntropy,
    'log_likelihood': CrossEntropy,
    'log likelihood': CrossEntropy,
    'nll': CrossEntropy,
    'negative_log_likelihood': CrossEntropy,
    'negative log likelihood': CrossEntropy,
    # Reversed CrossEntropy
    'rce': ReversedCrossEntropy,
    'reversed_crossentropy': ReversedCrossEntropy,
    'reversed crossentropy': ReversedCrossEntropy,
    'rbce': ReversedCrossEntropy,
    'reversed_binary_crossentropy': ReversedCrossEntropy,
    'reversed binary crossentropy': ReversedCrossEntropy,
    'rcce': ReversedCrossEntropy,
    'reversed_categorical_crossentropy': ReversedCrossEntropy,
    'reversed categorical crossentropy': ReversedCrossEntropy,
    'rll': ReversedCrossEntropy,
    'reversed_log_likelihood': ReversedCrossEntropy,
    'reversed log likelihood': ReversedCrossEntropy,
    'rnll': ReversedCrossEntropy,
    'reversed_negative_log_likelihood': ReversedCrossEntropy,
    'reversed negative log likelihood': ReversedCrossEntropy,
    # Symmetric CrossEntropy
    'sce': SymmetricCrossEntropy,
    'symmetric_crossentropy': SymmetricCrossEntropy,
    'symmetric crossentropy': SymmetricCrossEntropy,
    'sbce': SymmetricCrossEntropy,
    'symmetric_binary_crossentropy': SymmetricCrossEntropy,
    'symmetric binary crossentropy': SymmetricCrossEntropy,
    'scce': SymmetricCrossEntropy,
    'symmetric_categorical_crossentropy': SymmetricCrossEntropy,
    'symmetric categorical crossentropy': SymmetricCrossEntropy,
    'sll': SymmetricCrossEntropy,
    'symmetric_log_likelihood': SymmetricCrossEntropy,
    'symmetric log likelihood': SymmetricCrossEntropy,
    'snll': SymmetricCrossEntropy,
    'symmetric_negative_log_likelihood': SymmetricCrossEntropy,
    'symmetric negative log likelihood': SymmetricCrossEntropy
}
"""Dictionary which associates to each loss alias the respective class type."""
